from featureflow import Node, Feature, Decoder
from timeseries import \
    ConstantRateTimeSeries, ConstantRateTimeSeriesMetadata
from samplerate import AudioSampleRate, audio_sample_rate
import numpy as np
from duration import Seconds


class AudioSamples(ConstantRateTimeSeries):
    def __new__(cls, array, samplerate):
        if not isinstance(samplerate, AudioSampleRate):
            raise TypeError('samplerate should be an AudioSampleRate instance')
        return ConstantRateTimeSeries.__new__(
                cls, array, samplerate.frequency, samplerate.duration)

    @classmethod
    def from_example(cls, arr, example):
        return cls(arr, example.samplerate)

    @property
    def channels(self):
        if len(self.shape) == 1:
            return 1
        return self.shape[1]

    @property
    def samplerate(self):
        return audio_sample_rate(self.samples_per_second)

    @property
    def mono(self):
        if self.channels == 1:
            return self
        return AudioSamples(self.sum(axis=1) * 0.5, self.samplerate)


class AudioSamplesEncoder(Node):
    content_type = 'application/octet-stream'

    def __init__(self, needs=None):
        super(AudioSamplesEncoder, self).__init__(needs=needs)
        self.metadata = None

    def _process(self, data):
        if not self.metadata:
            self.metadata = ConstantRateTimeSeriesMetadata \
                .from_timeseries(data)
            packed = self.metadata.pack()
            yield packed

        encoded = data.tostring()
        yield encoded


def _np_from_buffer(b, shape, dtype, freq, duration):
    f = np.frombuffer if len(b) else np.fromstring
    shape = tuple(int(x) for x in shape)
    f = f(b, dtype=dtype).reshape(shape)
    samples_per_second = Seconds(1) / freq
    samplerate = audio_sample_rate(int(samples_per_second))
    return AudioSamples(f, samplerate)


class GreedyAudioSamplesDecoder(Decoder):
    def __init__(self):
        super(GreedyAudioSamplesDecoder, self).__init__()

    def __call__(self, flo):
        metadata, bytes_read = ConstantRateTimeSeriesMetadata.unpack(flo)

        leftovers = flo.read()
        leftover_bytes = len(leftovers)
        first_dim = leftover_bytes / metadata.totalsize
        dim = (first_dim,) + metadata.shape
        out = _np_from_buffer(
                leftovers,
                dim,
                metadata.dtype,
                metadata.frequency,
                metadata.duration)
        return out

    def __iter__(self, flo):
        yield self(flo)


class AudioSamplesFeature(Feature):
    def __init__(
            self,
            extractor,
            needs=None,
            store=False,
            key=None,
            encoder=AudioSamplesEncoder,
            decoder=GreedyAudioSamplesDecoder(),
            **extractor_args):
        super(AudioSamplesFeature, self).__init__(
                extractor,
                needs=needs,
                store=store,
                encoder=encoder,
                decoder=decoder,
                key=key,
                **extractor_args)



from featureflow import Node, Aggregator, BaseModel, Feature, PickleFeature
import numpy as np
from bisect import bisect_left
from zounds.timeseries import \
    TimeSlice, ConstantRateTimeSeriesFeature, ConstantRateTimeSeriesEncoder, \
    PackedConstantRateTimeSeriesEncoder
from zounds.nputil import packed_hamming_distance


class Index(Node):
    def __init__(self, func=None, needs=None):
        super(Index, self).__init__(needs=needs)
        self._func = func

    def _process(self, data):
        for _id in data.iter_ids():
            print _id
            yield _id, self._func(_id)


class Contiguous(Node):
    def __init__(self, needs=None):
        super(Contiguous, self).__init__(needs=needs)

    def _process(self, data):
        _id, data = data
        yield data


class Offsets(Aggregator, Node):
    def __init__(self, needs=None):
        super(Offsets, self).__init__(needs=needs)
        self._cache = ([], [])
        self._offset = 0

    def _enqueue(self, data, pusher):
        _id, data = data
        self._cache[0].append(_id)
        self._cache[1].append(self._offset)
        self._offset += len(data)


class SearchResults(object):
    def __init__(self, contiguous, offsets, indices, query):
        super(SearchResults, self).__init__()
        self._contiguous = contiguous
        self._offsets = offsets
        self._indices = indices
        self._query = query

    @staticmethod
    def _bisect(l, x):
        i = bisect_left(l, x)
        if i == len(l):
            return len(l) - 1
        if x < l[i]:
            i -= 1
        return i

    def __iter__(self):
        _ids, positions = self._offsets
        for i in self._indices:
            start_index = self._bisect(positions, i)
            diff = i - positions[start_index]
            _id = _ids[start_index]
            start_time = self._contiguous.frequency * diff
            duration = self._contiguous.duration
            ts = TimeSlice(duration, start_time)
            yield _id, ts


class Search(object):
    def __init__(self, contiguous, offsets):
        super(Search, self).__init__()
        self._contiguous = contiguous
        self._offsets = offsets

    def _score(self, query):
        raise NotImplementedError()

    # def _bisect(self, l, x):
    #     i = bisect_left(l, x)
    #     if i == len(l):
    #         return len(l) - 1
    #     if x < l[i]:
    #         i -= 1
    #     return i

    def search(self, query, n_results=5):
        scores = self._score(query)
        indices = np.argsort(scores)[:n_results]
        return SearchResults(self._contiguous, self._offsets, indices, query)
        # _ids, positions = self._offsets
        # for i in indices:
        #     start_index = self._bisect(positions, i)
        #     diff = i - positions[start_index]
        #     _id = _ids[start_index]
        #     start_time = self._contiguous.frequency * diff
        #     duration = self._contiguous.duration
        #     ts = TimeSlice(duration, start_time)
        #     yield _id, ts


class HammingDistanceSearch(Search):
    def __init__(self, contiguous, offsets):
        super(HammingDistanceSearch, self).__init__(contiguous, offsets)

    def _score(self, query):
        return np.logical_xor(query, self._contiguous).sum(axis=1)


class PackedHammingDistanceSearch(Search):
    def __init__(self, contiguous, offsets):
        super(PackedHammingDistanceSearch, self).__init__(contiguous, offsets)

    def _score(self, query):
        return packed_hamming_distance( \
                query.view(np.uint64), self._contiguous.view(np.uint64))


def hamming_distance_index(feature_func, packed):
    encoder = \
        PackedConstantRateTimeSeriesEncoder if packed \
            else ConstantRateTimeSeriesEncoder

    class Index(BaseModel):
        index = Feature( \
                Index,
                func=feature_func,
                store=False)

        contiguous = ConstantRateTimeSeriesFeature( \
                Contiguous,
                needs=index,
                encoder=encoder,
                store=True)

        offsets = PickleFeature( \
                Offsets,
                needs=index,
                store=True)

    return Index



# from __future__ import division
#
# import os
# from io import BytesIO
# from random import choice
# from uuid import uuid4
#
# import numpy as np
# import unittest2
# from soundfile import *
#
# from zounds.soundfile import AudioStream, OggVorbis, Resampler
# from zounds.timeseries import SR44100
# from featureflow import \
#     BaseModel, ByteStream, ByteStreamFeature, Feature, UuidProvider, \
#     InMemoryDatabase, StringDelimitedKeyBuilder, PersistenceSettings
# from featureflow.nmpy import NumpyFeature
#
# _sample_rates = (11025, 22050, 44100, 48000, 88200, 96000)
# _channels = (1, 2)
# _formats = (
#     ('WAV', 'PCM_16'),
#     ('WAV', 'PCM_24'),
#     ('WAV', 'PCM_32'),
#     ('WAV', 'FLOAT'),
#     ('WAV', 'DOUBLE'),
#
#     ('AIFF', 'PCM_16'),
#     ('AIFF', 'PCM_24'),
#     ('AIFF', 'PCM_32'),
#     ('AIFF', 'FLOAT'),
#     ('AIFF', 'DOUBLE'),
#
#     ('FLAC', 'PCM_16'),
#     ('FLAC', 'PCM_24'),
#
#     ('OGG', 'VORBIS')
# )
#
#
# class FuzzTests(unittest2.TestCase):
#     def __init__(
#             self, chunksize_bytes, samplerate, fmt, subtype, channels, seconds):
#
#         super(FuzzTests, self).__init__()
#         self._samplerate = samplerate
#         self._fmt = fmt
#         self._subtype = subtype
#         self._seconds = seconds
#         self._channels = channels
#         self._chunksize_bytes = chunksize_bytes
#
#     def __repr__(self):
#         return 'FuzzTests(sr = {_samplerate}, fmt = {_fmt}, st = {_subtype}, secs = {_seconds}, ch = {_channels}, cs = {_chunksize_bytes})'.format(
#                 **self.__dict__)
#
#     def __str__(self):
#         return self.__repr__()
#
#     def model_cls(self):
#
#         class Settings(PersistenceSettings):
#             id_provider = UuidProvider()
#             key_builder = StringDelimitedKeyBuilder()
#             database = InMemoryDatabase(key_builder=key_builder)
#
#         class Document(BaseModel, Settings):
#             raw = ByteStreamFeature(
#                     ByteStream,
#                     chunksize=self._chunksize_bytes,
#                     store=False)
#
#             ogg = Feature(
#                     OggVorbis,
#                     needs=raw,
#                     store=True)
#
#             pcm = NumpyFeature(
#                     AudioStream,
#                     needs=raw,
#                     store=True)
#
#             resampled = NumpyFeature(
#                     Resampler,
#                     samplerate=SR44100(),
#                     needs=pcm,
#                     store=True)
#
#         return Document
#
#     def runTest(self):
#         # TODO: Update this to use BytesIO instead of writing a file to disk
#         self._fn = '/tmp/' + uuid4().hex
#         print self
#
#         n_samples = int(self._samplerate * self._seconds)
#         samples = np.sin(np.arange(0, n_samples * 440, 440) * (2 * np.pi))
#         if self._channels == 2:
#             samples = np.repeat(samples, 2).reshape((n_samples, 2))
#
#         try:
#             with SoundFile(
#                     self._fn,
#                     mode='w',
#                     samplerate=self._samplerate,
#                     channels=self._channels,
#                     format=self._fmt,
#                     subtype=self._subtype) as sf:
#                 for i in range(0, n_samples, 44100):
#                     sf.write(samples[i: i + 44100])
#         except ValueError as e:
#             self.fail(e)
#
#         class HasUri(object):
#             def __init__(self, uri):
#                 self.uri = uri
#
#         model = self.model_cls()
#         _id = model.process(raw=HasUri(self._fn))
#         doc = model(_id)
#         orig_samples = doc.pcm
#         self.assertAlmostEqual(
#                 samples.shape[0], orig_samples.shape[0], delta=1)
#
#         del orig_samples
#         resampled = doc.resampled
#         seconds = resampled.shape[0] / 44100
#         self.assertAlmostEqual(self._seconds, seconds, delta=.025)
#         del resampled
#         ogg_bytes = doc.ogg
#
#         # first, do the ogg conversion "by hand" to make sure I'm not missing
#         # something
#         bio = BytesIO()
#         with SoundFile( \
#                 bio,
#                 format='OGG',
#                 subtype='VORBIS',
#                 mode='w',
#                 samplerate=self._samplerate,
#                 channels=self._channels) as ogg_sf:
#             for i in xrange(0, n_samples, 44100):
#                 ogg_sf.write(samples[i: i + 44100])
#
#         bio.seek(0)
#         ogg_bytes.seek(0)
#         bio.seek(0)
#
#         with SoundFile(ogg_bytes) as ogg_sf:
#             ogg_samples = ogg_sf.read(samples.shape[0] + 99999)
#
#         ogg_seconds = ogg_samples.shape[0] / self._samplerate
#         self.assertAlmostEqual(self._seconds, ogg_seconds, delta=.025)
#
#     def tearDown(self):
#         os.remove(self._fn)
#
#
# def suite():
#     suite = unittest2.TestSuite()
#
#     for _ in xrange(50):
#         seconds = (np.random.random_sample() * 50)
#         min_size = 4 * 96000 * 5 * 2
#         chunksize = min_size + (np.random.randint(0, 4 * 96000 * 25 * 2))
#         samplerate = choice(_sample_rates)
#         fmt = choice(_formats)
#         channels = choice(_channels)
#         suite.addTest(
#             FuzzTests(chunksize, samplerate, fmt[0], fmt[1], channels,
#                       seconds))
#
#     return suite
#
#
# if __name__ == '__main__':
#     unittest2.TextTestRunner().run(suite())



from featureflow import Node, Aggregator
import os
from soundfile import SoundFile
import requests
import json


class AudioMetaData(object):
    def __init__(
            self,
            uri=None,
            samplerate=None,
            channels=None,
            licensing=None,
            description=None,
            tags=None):
        super(AudioMetaData, self).__init__()
        self.uri = uri
        self.samplerate = samplerate
        self.channels = channels
        self.licensing = licensing
        self.description = description
        self.tags = tags


class AudioMetaDataEncoder(Aggregator, Node):
    content_type = 'application/json'

    def __init__(self, needs=None):
        super(AudioMetaDataEncoder, self).__init__(needs=needs)

    def _process(self, data):
        yield json.dumps({
            'uri': data.uri.url \
                if isinstance(data.uri, requests.Request) else data.uri,
            'samplerate': data.samplerate,
            'channels': data.channels,
            'licensing': data.licensing,
            'description': data.description,
            'tags': data.tags
        })


class MetaData(Node):
    def __init__(
            self,
            needs=None,
            freesound_api_key=None):

        super(MetaData, self).__init__(needs=needs)
        self._freesound_api_key = freesound_api_key

    def _ensure_freesound_api_key(self):
        if self._freesound_api_key is not None:
            return
        msg = \
            'A freesound uri was requested, but no freesound API key was provided'
        raise Exception(msg)

    def _process(self, data):
        uri = data
        if isinstance(data, requests.Request):
            yield AudioMetaData(uri=uri)
        elif os.path.exists(uri):
            sf = SoundFile(uri)
            yield AudioMetaData(
                    uri=uri,
                    samplerate=sf.samplerate,
                    channels=sf.channels)
        elif 'freesound.org' in uri:
            self._ensure_freesound_api_key()
            params = {'api_key': self._freesound_api_key}
            meta = requests.get(
                    uri, params=params).json()
            req = requests.Request(
                    'GET', meta['serve'], params=params)
            yield AudioMetaData(
                    uri=req,
                    samplerate=meta['samplerate'],
                    channels=meta['channels'],
                    licensing=meta['license'],
                    description=meta['description'],
                    tags=meta['tags'])
        else:
            raise Exception('Cannot handle uri {uri}'.format(**locals()))



from __future__ import division
import numpy as np
from scipy.fftpack import idct
from zounds.timeseries import \
    audio_sample_rate, AudioSamples, ConstantRateTimeSeries


class ShortTimeTransformSynthesizer(object):
    def __init__(self):
        super(ShortTimeTransformSynthesizer, self).__init__()

    def _transform(self, frames):
        return frames

    def _overlap_add(self, frames):
        sample_length_seconds = frames.duration_in_seconds / frames.shape[-1]
        samples_per_second = int(1 / sample_length_seconds)
        samplerate = audio_sample_rate(samples_per_second)
        windowsize = int(frames.duration / samplerate.frequency)
        hopsize = int(frames.frequency / samplerate.frequency)
        arr = np.zeros(frames.end / samplerate.frequency)
        for i, f in enumerate(frames):
            start = i * hopsize
            stop = start + windowsize
            arr[start:stop] += f
        return AudioSamples(arr, samplerate)

    def synthesize(self, frames):
        audio = self._transform(frames)
        ts = ConstantRateTimeSeries.from_example(audio, frames)
        return self._overlap_add(ts)


class FFTSynthesizer(ShortTimeTransformSynthesizer):
    def __init__(self):
        super(FFTSynthesizer, self).__init__()

    def _transform(self, frames):
        return np.fft.fftpack.ifft(frames)


class DCTSynthesizer(ShortTimeTransformSynthesizer):
    def __init__(self):
        super(DCTSynthesizer, self).__init__()

    def _transform(self, frames):
        return idct(frames, norm='ortho')



from featureflow import Node, NotEnoughData
import numpy as np
from zounds.nputil import windowed, sliding_window
from zounds.timeseries import ConstantRateTimeSeries


def oggvorbis(s):
    '''
    This is taken from the ogg vorbis spec 
    (http://xiph.org/vorbis/doc/Vorbis_I_spec.html)

    s is the total length of the window, in samples
    '''
    try:
        s = np.arange(s)
    except TypeError:
        s = np.arange(s[0])

    i = np.sin((s + .5) / len(s) * np.pi) ** 2
    f = np.sin(.5 * np.pi * i)
    return f * (1. / f.max())


class WindowingFunc(object):
    def _wdata(self, size):
        return np.ones(size)

    def __mul__(self, other):
        size = other.shape[1:]
        return self._wdata(size) * other

    def __rmul__(self, other):
        return self.__mul__(other)


class IdentityWindowingFunc(WindowingFunc):
    def __init__(self):
        super(IdentityWindowingFunc, self).__init__()


class OggVorbisWindowingFunc(WindowingFunc):
    def __init__(self):
        super(OggVorbisWindowingFunc, self).__init__()

    def _wdata(self, size):
        return oggvorbis(size)


class SlidingWindow(Node):
    def __init__(self, wscheme, wfunc=None, padwith=0, needs=None):
        super(SlidingWindow, self).__init__(needs=needs)
        self._scheme = wscheme
        self._func = wfunc or IdentityWindowingFunc()
        self._padwith = padwith
        self._cache = None

    def _first_chunk(self, data):
        padding = np.zeros((self._padwith,) + data.shape[1:], dtype=data.dtype)
        padding_ts = ConstantRateTimeSeries( \
                padding,
                data.frequency,
                data.duration)
        return padding_ts.concatenate(data)

    def _enqueue(self, data, pusher):
        if self._cache is None:
            self._cache = data
            # BUG: I Think this may only work in cases where frequency and
            # duration are the same
            self._windowsize = \
                int((self._scheme.duration - data.overlap) / data.frequency)
            self._stepsize = int(self._scheme.frequency / data.frequency)
        else:
            self._cache = np.concatenate([self._cache, data])

    def _dequeue(self):
        leftover, arr = windowed( \
                self._cache,
                self._windowsize,
                self._stepsize,
                dopad=self._finalized)

        self._cache = leftover

        if not arr.size:
            raise NotEnoughData()

        # BUG: Order matters here (try arr * self._func instead)
        # why does that statement result in __rmul__ being called for each
        # scalar value in arr?
        out = (self._func * arr) if self._func else arr
        out = ConstantRateTimeSeries( \
                out, self._scheme.frequency, self._scheme.duration)
        return out


# KLUDGE: This extractor works when trying to get random patches over the whole
# database, but doesn't work if we were trying to get a sliding window over, say,
# the bark bands of a single sound.  For that to work, we'd have to implement
# enqueue and dequeue methods that keep leftovers around between calls.
#
# If we're processing bark bands of a single sound all at once, we expect the
# sliding window to move over the time dimension first, and then over the frequency
# dimension.  Since each incoming chunk is of inderteminate size, we're left with
# nonsense ouput. This problem could be addressed by transposing the incoming data,
# so that the frequency dimension is traversed first, and then the time dimension.
class NDSlidingWindow(Node):
    def __init__(self, windowsize=None, stepsize=None, needs=None):
        super(NDSlidingWindow, self).__init__(needs=needs)
        self._ws = windowsize
        self._ss = stepsize

    def _process(self, data):
        try:
            yield sliding_window(data, self._ws, self._ss)
        except ValueError:
            pass



from learn import KMeans, BinaryRbm, LinearRbm, Learned

from preprocess import \
    MeanStdNormalization, UnitNorm, Log, PreprocessingPipeline

from random_samples import ReservoirSampler

from template_match import TemplateMatch


import unittest2
import numpy as np
from synthesize import ShortTimeTransformSynthesizer
from zounds.timeseries import \
    ConstantRateTimeSeries, SR44100, HalfLapped, AudioSamples


class SynthesizeTests(unittest2.TestCase):
    def test_has_correct_sample_rate(self):
        half_lapped = HalfLapped()
        synth = ShortTimeTransformSynthesizer()
        raw = np.zeros((100, 2048))
        timeseries = ConstantRateTimeSeries(
                raw, half_lapped.frequency, half_lapped.duration)
        output = synth.synthesize(timeseries)
        self.assertIsInstance(output.samplerate, SR44100)
        self.assertIsInstance(output, AudioSamples)



from __future__ import division

from io import BytesIO

import numpy as np
import unittest2
from soundfile import SoundFile

from zounds.timeseries import \
    TimeSlice, ConstantRateTimeSeriesFeature, AudioSamples, AudioSamplesFeature, \
    SR44100, HalfLapped, Seconds, Milliseconds
from zounds.soundfile import \
    AudioStream, OggVorbis, OggVorbisFeature, Resampler
from zounds.spectral import \
    SlidingWindow, OggVorbisWindowingFunc, FFT, Chroma, BarkBands, BFCC
from zounds.basic import Max
from featureflow import *

windowing_scheme = HalfLapped()
samplerate = SR44100()


class Settings(PersistenceSettings):
    id_provider = UuidProvider()
    key_builder = StringDelimitedKeyBuilder()
    database = InMemoryDatabase(key_builder=key_builder)


class Document(BaseModel, Settings):
    raw = ByteStreamFeature(
            ByteStream,
            chunksize=2 * 44100 * 30 * 2,
            store=True)

    ogg = OggVorbisFeature(
            OggVorbis,
            needs=raw,
            store=True)

    pcm = AudioSamplesFeature(
            AudioStream,
            needs=raw,
            store=True)

    resampled = AudioSamplesFeature(
            Resampler,
            needs=pcm,
            samplerate=samplerate,
            store=True)

    windowed = ConstantRateTimeSeriesFeature(
            SlidingWindow,
            needs=resampled,
            wscheme=windowing_scheme,
            wfunc=OggVorbisWindowingFunc(),
            store=False)

    fft = ConstantRateTimeSeriesFeature(
            FFT,
            needs=windowed,
            store=True)

    chroma = ConstantRateTimeSeriesFeature(
            Chroma,
            needs=fft,
            samplerate=samplerate,
            store=True)

    bark = ConstantRateTimeSeriesFeature(
            BarkBands,
            needs=fft,
            samplerate=samplerate,
            store=True)

    bfcc = ConstantRateTimeSeriesFeature(
            BFCC,
            needs=bark,
            store=True)

    bfcc_sliding_window = ConstantRateTimeSeriesFeature(
            SlidingWindow,
            needs=bfcc,
            wscheme=windowing_scheme * (2, 4),
            store=True)

    bfcc_pooled = ConstantRateTimeSeriesFeature(
            Max,
            needs=bfcc_sliding_window,
            axis=1,
            store=True)


class HasUri(object):
    def __init__(self, uri):
        super(HasUri, self).__init__()
        self.uri = uri


def signal(hz=440, seconds=5., sr=44100.):
    # cycles per sample
    cps = hz / sr
    # total samples
    ts = seconds * sr
    mono = np.sin(np.arange(0, ts * cps, cps) * (2 * np.pi))
    return np.column_stack((mono, mono))


def soundfile(hz=440, seconds=5., sr=44100.):
    bio = BytesIO()
    s = signal(hz, seconds, sr)
    with SoundFile(
            bio,
            mode='w',
            channels=2,
            format='WAV',
            subtype='PCM_16',
            samplerate=int(sr)) as f:
        f.write(s)
    bio.seek(0)
    return s, HasUri(bio)


class IntegrationTests(unittest2.TestCase):
    def setUp(self):
        signal, bio = soundfile(seconds=10.)
        _id = Document.process(raw=bio)
        self.doc = Document(_id)
        self.signal = signal

    def test_pcm_returns_audio_samples(self):
        self.assertIsInstance(self.doc.pcm, AudioSamples)

    def test_pcm_is_summed_to_mono(self):
        self.assertEqual(1, self.doc.pcm.channels)

    def test_resampled_returns_audio_samples(self):
        self.assertIsInstance(self.doc.resampled, AudioSamples)

    def test_ogg_vorbis_iter_chunks_returns_audio_samples(self):
        chunks = list(self.doc.ogg.iter_chunks())
        self.assertTrue(
                all(isinstance(chunk, AudioSamples) for chunk in chunks))

    def test_ogg_vorbis_wrapper_returns_audio_samples(self):
        self.assertIsInstance(self.doc.ogg[:], AudioSamples)

    def test_ogg_wrapper_has_correct_duration_seconds(self):
        self.assertEqual(10, self.doc.ogg.duration_seconds)

    def test_windowed_and_fft_have_same_first_dimension(self):
        self.assertEqual(self.doc.windowed.shape[0], self.doc.fft.shape[0])

    def test_fft_dimension_is_half_of_windowsize(self):
        self.assertEqual(self.doc.windowed.shape[1] // 2, self.doc.fft.shape[1])

    def test_bfcc_sliding_window_has_correct_shape(self):
        self.assertEqual((4, 13), self.doc.bfcc_sliding_window.shape[1:])

    def test_bfcc_sliding_window_has_correct_frequency(self):
        self.assertEqual(
                2,
                self.doc.bfcc_sliding_window.frequency / self.doc.bfcc.frequency)

    def test_bfcc_sliding_window_has_correct_duration(self):
        self.assertEqual(
                5,
                self.doc.bfcc_sliding_window.duration / self.doc.bfcc.frequency)

    def test_bfcc_pooled_has_correct_shape(self):
        self.assertEqual(2, len(self.doc.bfcc_pooled.shape))
        self.assertEqual((13,), self.doc.bfcc_pooled.shape[1:])

    def test_bfcc_pooled_has_correct_frequency(self):
        self.assertEqual(
                2, self.doc.bfcc_pooled.frequency / self.doc.bfcc.frequency)

    def test_bfcc_pooled_has_correct_duration(self):
        self.assertEqual(
                5, self.doc.bfcc_pooled.duration / self.doc.bfcc.frequency)

    def test_can_get_second_long_slice_from_ogg_vorbis_feature(self):
        ogg = self.doc.ogg
        samples = ogg[TimeSlice(Seconds(1), start=Seconds(5))]
        self.assertEqual(44100, len(samples))

    def test_can_get_entirety_of_ogg_vorbis_feature_with_slice(self):
        ogg = self.doc.ogg
        samples = ogg[:]
        self.assertEqual(441000, len(samples))

    def test_can_read_ogg_samples_twice(self):
        ogg = self.doc.ogg
        s1 = ogg[:]
        s2 = ogg[:]
        self.assertEqual(s1.shape, s2.shape)

    def test_can_get_end_of_ogg_vorbis_feature_with_slice(self):
        ogg = self.doc.ogg
        samples = ogg[TimeSlice(Seconds(1), Milliseconds(9500))]
        self.assertEqual(22050, len(samples))

    def test_sliding_window_has_correct_relationship_to_bfcc(self):
        self.assertEqual(
                2,
                self.doc.bfcc.shape[0] // self.doc.bfcc_sliding_window.shape[0])



from __future__ import division
from abc import ABCMeta, abstractmethod
import numpy as np
from scipy.signal import triang


def fft_index(freq_hz, ws, sr, rnd=np.round):
    """
    Given a frequency in hz, a window size, and a sample rate,
    return the fft bin into which the freq in hz falls
    """
    if freq_hz < 0 or freq_hz > sr / 2.:
        raise ValueError( \
                'Freq must be greater than zero and less than the Nyquist frequency')

    fft_bandwidth = (sr * .5) / (ws * .5)
    return int(rnd(freq_hz / fft_bandwidth))


def fft_span(start_hz, stop_hz, ws, sr):
    """
    Given a span in hz, return the start and stop fft bin indices covering
    that span
    """
    s_index = fft_index(start_hz, ws, sr, rnd=np.floor)
    e_index = fft_index(stop_hz, ws, sr, rnd=np.ceil)
    # ensure that the span is at least one fft bin
    e_index = s_index + 1 if s_index == e_index else e_index
    return s_index, e_index


def _hz_is_valid(freq_hz, sr):
    if freq_hz < 0 or freq_hz > sr / 2.:
        raise ValueError( \
                'Freq must be greater than zero and less than the Nyquist frequency')


# TODO: Should this be different for Bark and Mel scales?  Isn't equivalent
# rectangular bandwidth based on the Bark scale?
def erb(hz):
    """
    equivalent rectangular bandwidth
    """
    return (0.108 * hz) + 24.7


# TODO: Documentation
class Scale(object):
    __metaclass__ = ABCMeta

    _data = dict()

    def __init__(self, samplerate, window_size, nbands, start_freq_hz,
                 stop_freq_hz):
        object.__init__(self)
        self._fft_bandwidth = (samplerate * .5) / (window_size * .5)
        self._sr = samplerate
        self._ws = window_size
        self._nb = nbands
        self._sfhz = start_freq_hz
        self._efhz = stop_freq_hz
        self._data_key = (self.__class__.__name__,
                          self._sr, self._ws, self._nb,
                          self._sfhz, self._efhz)

        if self._data_key not in self._data:
            self._data[self._data_key] = self._build_data()

    @abstractmethod
    def to_hz(self, s):
        pass

    @abstractmethod
    def from_hz(self, hz):
        pass

    def _build_data(self):
        start_unit = self.from_hz(self._sfhz)
        stop_unit = self.from_hz(self._efhz)
        bandwidth = (stop_unit - start_unit) / self._nb

        _slices = []
        _triwins = []

        for i in xrange(1, self._nb + 1):
            b = i * bandwidth
            hz = self.to_hz(b)
            _herb = erb(hz) / 2
            start_hz = hz - _herb
            start_hz = 0 if start_hz < 0 else start_hz
            stop_hz = hz + _herb
            _hz_is_valid(start_hz, self._sr)
            _hz_is_valid(stop_hz, self._sr)
            s = start_hz / self._fft_bandwidth
            e = stop_hz / self._fft_bandwidth
            s_index = int(np.floor(s))
            e_index = int(np.ceil(e))
            e_index = s_index + 1 if s_index == e_index else e_index
            n_bins = e_index - s_index
            _slices.append(slice(s_index, e_index))

            tri = triang(n_bins)
            peak_diff = (n_bins // 2) / (n_bins / 2)
            bin_diff = (e - s) / n_bins

            # scale the triangular window to account for aliasing introduced by
            # a) the rounded number of bins in the window
            # b) the fact that even numbered windows won't have a peak with value
            #    1 in the middle
            tri = tri * peak_diff * bin_diff
            _triwins.append(tri)

        return _slices, _triwins

    @property
    def n_bands(self):
        return self._nb

    @property
    def data(self):
        return self._data[self._data_key]

    def transform(self, fft):
        slices, triwins = self.data
        cb = np.ndarray((fft.shape[0], self._nb), dtype=np.float32)
        for i in xrange(self._nb):
            cb[:, i] = (fft[:, slices[i]] * triwins[i]).sum(1)
        return cb


class Bark(Scale):
    def __init__(self, samplerate, window_size, nbands, start_freq_hz,
                 stop_freq_hz):
        Scale.__init__( \
                self, samplerate, window_size, nbands, start_freq_hz,
                stop_freq_hz)

    def to_hz(self, b):
        return 300. * ((np.e ** (b / 6.0)) - (np.e ** (-b / 6.)))

    def from_hz(self, hz):
        return 6. * np.log((hz / 600.) + np.sqrt((hz / 600.) ** 2 + 1))


class Mel(Scale):
    def __init__(self, samplerate, window_size, nbands, start_freq_hz,
                 stop_freq_hz):
        Scale.__init__( \
                self, samplerate, window_size, nbands, start_freq_hz,
                stop_freq_hz)

    def to_hz(self, m):
        return 700 * ((np.e ** (m / 1127)) - 1)

    def from_hz(self, hz):
        return 1127 * np.log(1 + (hz / 700))


class Chroma(Scale):
    """
    Based on Dan Ellis' MATLAB code, available here:
    http://www.ee.columbia.edu/~dpwe/resources/matlab/chroma-ansyn/#1
    """

    def __init__(self, samplerate, window_size, nbands=12):
        # TODO: Set the start and stop freqs in hz
        self._a440 = 440.
        Scale.__init__( \
                self, samplerate, window_size, nbands, None, None)

    def to_hz(self):
        raise NotImplemented()

    def from_hz(self, hz):
        """
        Convert a frequency in hz into a real number representing the octaves
        above A0.
        """
        return np.log(hz / (self._a440 / 16)) / np.log(2)

    @property
    def n_fft_bands(self):
        return int(self._ws / 2.)

    # TODO: Plot each step to ensure that you understand what's going on
    def _build_data(self):
        # allocate memory for an n_chroma_bins x n_fft_bins matrix.  We'll be 
        # taking the dot product of this matrix with fft coefficients to get 
        # the chroma vector
        # weights = np.zeros((self.n_bands,self._ws))
        fft_bins = np.zeros(self._ws)

        # get hz values for every possible bin
        hz = (np.arange(1, self._ws) / self._ws) * self._sr

        # translate hz into real-valued octave numbers (distance from A0)
        fft_bins[1:] = self.n_bands * self.from_hz(hz)
        fft_bins[0] = fft_bins[1] - (1.5 * self.n_bands)

        # calculate the width, in octaves, of each bin
        bin_widths = np.concatenate([np.diff(fft_bins), [1]])

        # TODO: Is this tile necessary?
        a = np.tile(fft_bins, (self.n_bands, 1))
        # TODO: Is this tile necessary?
        b = np.tile(np.arange(self.n_bands), (self._ws, 1)).T
        # TODO: What is this variable for?
        D = a - b

        half_bands = np.round(self.n_bands / 2.)
        # I'm officially lost. back up and figure out what's happening
        D = ((D + half_bands + (10 * self.n_bands)) % self.n_bands) - half_bands

        weights = \
            np.exp(
                -0.5 * (2 * (D / np.tile(bin_widths, (self.n_bands, 1)))) ** 2)

        # normalize each column
        smsqr = np.sqrt((weights ** 2).sum(0))
        smsqr[smsqr == 0] = 1
        weights /= smsqr
        weights = weights[:, :self.n_fft_bands]
        return weights.T

    def transform(self, fft):
        return np.dot(fft, self.data)



from duration import \
    Hours, Minutes, Seconds, Milliseconds, Microseconds, Picoseconds

from audiosamples import \
    AudioSamples, AudioSamplesEncoder, GreedyAudioSamplesDecoder, \
    AudioSamplesFeature

from samplerate import \
    SR11025, SR22050, SR44100, SR48000, SR96000, HalfLapped, audio_sample_rate

from timeseries import \
    TimeSlice, ConstantRateTimeSeriesEncoder, ConstantRateTimeSeriesFeature, \
    GreedyConstantRateTimeSeriesDecoder, PackedConstantRateTimeSeriesEncoder, \
    ConstantRateTimeSeries



from io import BytesIO
from os import SEEK_END

from soundfile import SoundFile

from zounds.timeseries import AudioSamples, audio_sample_rate
from byte_depth import chunk_size_samples
from featureflow import Node


class AudioStream(Node):
    def __init__(
            self,
            sum_to_mono=True,
            needs=None):

        super(AudioStream, self).__init__(needs=needs)
        self._sum_to_mono = sum_to_mono
        self._buf = None
        self._sf = None
        self._chunk_size_samples = None
        self._cache = ''

    def _enqueue(self, data, pusher):
        self._cache += data

    def _dequeue(self):
        v = self._cache
        self._cache = ''
        return v

    def _get_samples(self):
        raw_samples = self._sf.read(self._chunk_size_samples)
        samples = AudioSamples(
                raw_samples, audio_sample_rate(self._sf.samplerate))
        if self._sum_to_mono:
            return samples.mono
        return samples

    def _process(self, data):
        b = data

        # TODO: Use the _first_chunk() hook instead of the if statement
        if self._buf is None:
            self._buf = MemoryBuffer(b.total_length)

        self._buf.write(b)

        if self._sf is None:
            self._sf = SoundFile(self._buf)

        if self._chunk_size_samples is None:
            self._chunk_size_samples = chunk_size_samples(self._sf, b)

        yield self._get_samples()

    def _last_chunk(self):
        samples = self._get_samples()
        while samples.size:
            yield samples
            samples = self._get_samples()


class MemoryBuffer(object):
    """
    This class is the implementation of the virtual io interface required by
    PySoundfile/libsndfile.

    Some of the stateful/hacky things in this class (see KLUDGE note below)
    could be avoided if this class had the following modifications:
        - it maintains its own position
        - it maintains the span of its buffer/BytesIO instance
        -
    """

    def __init__(self, content_length, max_size=10 * 1024 * 1024):
        super(MemoryBuffer, self).__init__()
        self._content_length = content_length
        self._buf = BytesIO()
        self._max_size = max_size
        self.tell = self._tell

        self._total_bytes_written = 0
        self._total_bytes_read = 0

    def read(self, count):
        if count == -1:
            return self._buf.read()
        data = self._buf.read(count)
        return data

    def readinto(self, buf):
        data = self.read(len(buf))
        ld = len(data)
        buf[:ld] = data
        return ld

    def write(self, data):
        read_pos = self._buf.tell()
        if read_pos > self._max_size:
            new_buf = BytesIO()
            new_buf.write(self._buf.read())
            self._buf = new_buf
            read_pos = 0
        self._buf.seek(0, 2)
        self._buf.write(data)
        self._buf.seek(read_pos)
        return len(data)

    def _tell(self):
        return self._buf.tell()

    def _tell_end(self):
        self.tell = self._tell
        return self._content_length

    def seek(self, offset, whence):
        if whence == SEEK_END:
            # KLUDGE: PySoundfile no longer supports __len__, which means that
            # this stateful garbage is required.
            self.tell = self._tell_end
        self._buf.seek(offset, whence)

    def flush(self):
        self._buf.flush()



from featureflow import Node
import numpy as np
from scipy.spatial.distance import cdist
from zounds.nputil import safe_unit_norm, sliding_window
from zounds.timeseries import ConstantRateTimeSeries


class TemplateMatch(Node):
    def __init__(self, templates=None, needs=None):
        super(TemplateMatch, self).__init__(needs=needs)
        self._windowsize = templates.shape[1:]
        twod_templates = templates.reshape(templates.shape[0], -1)
        self._templates = safe_unit_norm(twod_templates)

    def _process(self, data):
        out = np.zeros((data.shape[0], self._templates.shape[0]))
        for i, chunk in enumerate(data):
            patches = sliding_window( \
                    chunk,
                    self._windowsize,
                    (1,) * len(self._windowsize))
            twod_patches = patches.reshape(patches.shape[0], -1)
            normalized_patches = safe_unit_norm(twod_patches)
            dist = cdist(normalized_patches, self._templates).min(axis=0)
            dist[dist == 0] = 1e-2
            dist = 1. / dist
            out[i] = dist
        print out.shape
        yield ConstantRateTimeSeries(out, data.frequency, data.duration)



from __future__ import division

from ctypes import *

import numpy as np

from zounds.timeseries import AudioSamples, SR44100

libsamplerate = CDLL('libsamplerate.so')

from featureflow import Node


class SRC_DATA(Structure):
    '''
    A wrapper for the libsamplerate.SRC_DATA struct
    '''
    _fields_ = [('data_in', POINTER(c_float)),
                ('data_out', POINTER(c_float)),
                ('input_frames', c_long),
                ('output_frames', c_long),
                ('input_frames_used', c_long),
                ('output_frames_gen', c_long),
                ('end_of_input', c_int),
                ('src_ratio', c_double), ]


class Resample(object):
    """
    A wrapper around the libsamplerate src_process() method.  This class is
    intended for one-time use. New instances should be created for each sound\
    file processed.
    """

    def __init__(self, orig_sample_rate, new_sample_rate, \
                 nchannels=1, converter_type=1):
        """
        orig_sample_rate - The sample rate of the incoming samples, in hz
        new_sample_rate - The sample_rate of the outgoiing samples, in hz
        n_channels - Number of channels in the incoming and outgoing samples
        converter_type - See http://www.mega-nerd.com/SRC/api_misc.html#Converters
                         for a list of conversion types. "0" is the best-quality,
                         and slowest converter

        """
        self._ratio = new_sample_rate / orig_sample_rate
        # check if the conversion ratio is considered valid by libsamplerate
        if not libsamplerate.src_is_valid_ratio(c_double(self._ratio)):
            raise ValueError('%1.2f / %1.2f = %1.4f is not a valid ratio' % \
                             (new_sample_rate, orig_sample_rate, self._ratio))
        # create a pointer to the SRC_STATE struct, which maintains state
        # between calls to src_process()
        error = pointer(c_int(0))
        self.nchannels = nchannels
        self.converter_type = converter_type
        self._state = libsamplerate.src_new( \
                c_int(converter_type), c_int(nchannels), error)

    def _prepare_input(self, insamples):
        # ensure that the input is float data
        if np.float32 != insamples.dtype:
            insamples = insamples.astype(np.float32)

        return insamples

    def _output_buffer(self, insamples):
        outsize = int(np.round(insamples.size * self._ratio))
        return np.zeros(outsize, dtype=np.float32)

    def _src_data_struct(self, insamples, outsamples, end_of_input=False):
        # Build the SRC_DATA struct
        return SRC_DATA( \
                # a pointer to the input samples
                data_in=insamples.ctypes.data_as(POINTER(c_float)),
                # a pointer to the output buffer
                data_out=outsamples.ctypes.data_as(POINTER(c_float)),
                # number of input samples
                input_frames=insamples.size,
                # number of output samples
                output_frames=outsamples.size,
                # NOT the end of input, i.e., there is more data to process
                end_of_input=int(end_of_input),
                # the conversion ratio
                src_ratio=self._ratio)

    def _check_for_error(self, return_code):
        if return_code:
            raise Exception( \
                    c_char_p(
                        libsamplerate.src_strerror(c_int(return_code))).value)

    def all_at_once(self, insamples):
        insamples = self._prepare_input(insamples)
        outsamples = self._output_buffer(insamples)
        sd = self._src_data_struct(insamples, outsamples)
        rv = libsamplerate.src_simple(pointer(sd),
                                      c_int(self.converter_type),
                                      c_int(self.nchannels))
        self._check_for_error(rv)
        return outsamples

    def __call__(self, insamples, end_of_input=False):
        insamples = self._prepare_input(insamples)
        outsamples = self._output_buffer(insamples)
        sd = self._src_data_struct(insamples, outsamples, end_of_input)
        rv = libsamplerate.src_process(self._state, pointer(sd))
        self._check_for_error(rv)
        return outsamples


class Resampler(Node):
    def __init__(self, samplerate=None, needs=None):
        super(Resampler, self).__init__(needs=needs)
        self._samplerate = samplerate or SR44100()
        self._resample = None

    def _noop(self, data, finalized):
        return data

    def _process(self, data):
        sr = data.samples_per_second

        if self._resample is None:
            target_sr = self._samplerate.samples_per_second
            self._resample = Resample( \
                    sr,
                    target_sr,
                    1 if len(data.shape) == 1 else data.shape[1])

            if target_sr != sr:
                self._rs = self._resample
                # KLUDGE: The following line seems to solve a bug whereby 
                # libsamplerate doesn't generate enough samples the first time
                # src_process is called. We're calling it once here, so the "real"
                # output will come out click-free
                self._resample(np.zeros(target_sr, dtype=np.float32))
            else:
                self._rs = self._noop

        resampled = self._rs(data, self._finalized)
        if not isinstance(resampled, AudioSamples):
            resampled = AudioSamples( \
                    resampled, self._samplerate)
        yield resampled



from __future__ import division
from duration import Picoseconds


class SampleRate(object):
    def __init__(self, frequency, duration):
        self.frequency = frequency
        self.duration = duration
        super(SampleRate, self).__init__()

    @property
    def overlap(self):
        return self.duration - self.frequency

    def __mul__(self, other):
        try:
            if len(other) == 1:
                other *= 2
        except TypeError:
            other = (other, other)

        freq = self.frequency * other[0]
        duration = (self.frequency * other[1]) + self.overlap
        return SampleRate(freq, duration)


class AudioSampleRate(SampleRate):
    def __init__(self, samples_per_second):
        one_sample = Picoseconds(int(1e12)) // samples_per_second
        super(AudioSampleRate, self).__init__(one_sample, one_sample)

    @property
    def samples_per_second(self):
        return int(Picoseconds(int(1e12)) / self.frequency)


class SR96000(AudioSampleRate):
    def __init__(self):
        super(SR96000, self).__init__(96000)


class SR48000(AudioSampleRate):
    def __init__(self):
        super(SR48000, self).__init__(48000)


class SR44100(AudioSampleRate):
    def __init__(self):
        super(SR44100, self).__init__(44100)


class SR22050(AudioSampleRate):
    def __init__(self):
        super(SR22050, self).__init__(22050)


class SR11025(AudioSampleRate):
    def __init__(self):
        super(SR11025, self).__init__(11025)

_samplerates = (SR96000(), SR48000(), SR44100(), SR22050(), SR11025())


def audio_sample_rate(samples_per_second):
    for sr in _samplerates:
        if samples_per_second == sr.samples_per_second:
            return sr
    raise ValueError(
        '{samples_per_second} is an invalid sample rate'.format(**locals()))


class HalfLapped(SampleRate):
    def __init__(self, window_at_44100=2048, hop_at_44100=1024):
        one_sample_at_44100 = Picoseconds(int(1e12)) / 44100.
        window = one_sample_at_44100 * window_at_44100
        step = one_sample_at_44100 * hop_at_44100
        super(HalfLapped, self).__init__(step, window)





from __future__ import division
import numpy as np
from zounds.timeseries import ConstantRateTimeSeries, SR44100
from zounds.visualize import plot


class OctaveScale(object):
    def __init__(self, freq_min, freq_max, bands_per_octave):
        self.bands = int( \
                np.ceil(np.log2(freq_max / freq_min) * bands_per_octave)) + 1
        self.freq_min = freq_min
        self.freq_max = freq_max
        self.bands_per_octave = bands_per_octave
        self.pow2n = 2 ** (1 / self.bands_per_octave)
        self.quality_factor = np.sqrt(self.pow2n) / (self.pow2n - 1) / 2
        self.center_frequencies = \
            self.freq_min * self.pow2n ** np.arange(self.bands)


class Windows(object):
    def __init__(self, scale, timeseries):
        self.scale = scale
        self.samplerate = timeseries.samples_per_second
        print self.samplerate
        self.nsamples = len(timeseries)
        self.q = scale.quality_factor
        nyquist = self.samplerate / 2
        print nyquist
        cf = self.scale.center_frequencies
        # limit frequencies to those greater than zero, and less than the
        # nyquist frequency
        cf = cf[(cf > 0) & (cf <= nyquist)]

        q_needed = cf * (len(timeseries) / (8 * self.samplerate))
        if np.any(self.q > q_needed):
            raise Exception('Q factor too high')

        freqs = np.concatenate([(0,), cf, (nyquist,)])
        self.fbas = np.concatenate([freqs, self.samplerate - freqs[-2:0:-1]])
        self.fbas2 = self.fbas * (len(timeseries) / self.samplerate)


if __name__ == '__main__':
    scale = OctaveScale(20, 2e4, 12)
    sr = SR44100()
    ts = ConstantRateTimeSeries( \
            np.random.random_sample(44100 * 8), sr.frequency, sr.duration)
    windows = Windows(scale, ts)
    plot(windows.fbas, '/home/john/Desktop/fbas.png')
    plot(windows.fbas2, '/home/john/Desktop/fbas2.png')



from __future__ import division

_lookup = {
    # https://support.microsoft.com/en-us/kb/89879
    'MS_ADPCM': 1,
    'IMA_ADPCM': 1,
    'PCM_S8': 1,
    'PCM_U8': 1,
    'PCM_16': 2,
    'PCM_24': 3,
    'PCM_32': 4,
    'FLOAT': 4,
    'DOUBLE': 8,
    # this is a guess, erring on the side of caution
    'VORBIS': 3
}


def chunk_size_samples(sf, buf):
    """
    Black magic to account for the fact that libsndfile's behavior varies
    depending on file format when using the virtual io api.

    If you ask for more samples from an ogg or flac file than are available
    at that moment, libsndfile will give you no more samples ever, even if
    more bytes arrive in the buffer later.
    """
    byte_depth = _lookup[sf.subtype]
    channels = sf.channels
    bytes_per_second = byte_depth * sf.samplerate * channels
    secs = len(buf) / bytes_per_second
    secs = max(1, secs - 6)
    return int(secs * sf.samplerate)



from audio_metadata import MetaData, AudioMetaDataEncoder

from ogg_vorbis import \
    OggVorbis, OggVorbisDecoder, OggVorbisEncoder, OggVorbisFeature, \
    OggVorbisWrapper

from audiostream import AudioStream

from resample import Resampler



__version__ = '0.1.1'

from timeseries import \
    Hours, Minutes, Seconds, Milliseconds, Microseconds, Picoseconds, \
    AudioSamples, AudioSamplesEncoder, GreedyAudioSamplesDecoder, \
    AudioSamplesFeature, \
    SR11025, SR22050, SR44100, SR48000, SR96000, HalfLapped, \
    TimeSlice, ConstantRateTimeSeriesEncoder, ConstantRateTimeSeriesFeature, \
    GreedyConstantRateTimeSeriesDecoder, PackedConstantRateTimeSeriesEncoder

from soundfile import \
    MetaData, AudioMetaDataEncoder, \
    OggVorbis, OggVorbisDecoder, OggVorbisEncoder, OggVorbisFeature, \
    OggVorbisWrapper, \
    AudioStream, \
    Resampler

from spectral import \
    SlidingWindow, OggVorbisWindowingFunc, \
    FFT, DCT, BarkBands, Chroma, BFCC, SpectralCentroid, SpectralFlatness

from segment import \
    MeasureOfTransience, MovingAveragePeakPicker, SparseTimestampDecoder, \
    SparseTimestampEncoder, TimeSliceDecoder, TimeSliceFeature, ComplexDomain

from synthesize import FFTSynthesizer, DCTSynthesizer

from learn import \
    KMeans, BinaryRbm, LinearRbm, Learned, \
    MeanStdNormalization, UnitNorm, Log, PreprocessingPipeline, \
    ReservoirSampler, \
    TemplateMatch

from ui import ZoundsApp, RangeUnitUnsupportedException

from index import \
    Index, Offsets, Contiguous, HammingDistanceSearch, \
    PackedHammingDistanceSearch, SearchResults

from basic import Slice, Sum, Max, process_dir



import os
import sys
import json
import tornado.ioloop
import tornado.web
import httplib
import numpy as np
from featureflow import Decoder
import traceback
from matplotlib import pyplot as plt
from io import BytesIO
from cStringIO import StringIO
import ast
import re
import uuid
import datetime
from zounds.timeseries import \
    ConstantRateTimeSeriesFeature, AudioSamples, Seconds, Picoseconds, TimeSlice
from zounds.segment import TimeSliceFeature
from zounds.index import SearchResults
from zounds.soundfile import OggVorbisFeature
from soundfile import SoundFile


class RangeUnitUnsupportedException(Exception):
    """
    Raised when an HTTP range request is made with a unit not supported by this
    application
    """
    pass


class UnsatisfiableRangeRequestException(Exception):
    """
    Exception raised when an HTTP range request cannot be satisfied for a
    particular resource
    """
    pass


class RangeRequest(object):
    def __init__(self, range_header):
        self.range_header = range_header
        self.re = re.compile(
                r'^(?P<unit>[^=]+)=(?P<start>[^-]+)-(?P<stop>.*?)$')

    def time_slice(self, start, stop):
        start = float(start)
        try:
            stop = float(stop)
        except ValueError:
            stop = None
        duration = \
            None if stop is None else Picoseconds(int(1e12 * (stop - start)))
        start = Picoseconds(int(1e12 * start))
        return TimeSlice(duration, start=start)

    def byte_slice(self, start, stop):
        start = int(start)
        try:
            stop = int(stop)
        except ValueError:
            stop = None
        return slice(start, stop)

    def range(self):
        raw = self.range_header
        if not raw:
            return slice(None)

        m = self.re.match(raw)
        if not m:
            return slice(None)

        units = m.groupdict()['unit']
        start = m.groupdict()['start']
        stop = m.groupdict()['stop']

        if units == 'bytes':
            return self.byte_slice(start, stop)
        elif units == 'seconds':
            return self.time_slice(start, stop)
        else:
            raise RangeUnitUnsupportedException(units)


class NoMatchingSerializerException(Exception):
    pass


class ContentRange(object):
    def __init__(self, unit, start, total, stop=None):
        self.unit = unit
        self.total = total
        self.stop = stop
        self.start = start

    @staticmethod
    def from_timeslice(timeslice, total):
        one_second = Seconds(1)
        stop = None
        start = timeslice.start / one_second
        if timeslice.duration is not None:
            stop = start + (timeslice.duration / one_second)
        return ContentRange(
                'seconds',
                timeslice.start / one_second,
                total / one_second,
                stop)

    @staticmethod
    def from_slice(slce, total):
        return ContentRange('bytes', slce.start, total, slce.stop)

    def __str__(self):
        unit = self.unit
        start = self.start
        stop = self.stop or self.total
        total = self.total
        return '{unit} {start}-{stop}/{total}'.format(**locals())


class TempResult(object):
    def __init__(
            self,
            data,
            content_type,
            is_partial=False,
            content_range=None):
        self.content_range = content_range
        self.data = data
        self.content_type = content_type
        self.timestamp = datetime.datetime.utcnow()
        self.is_partial = is_partial


class RequestContext(object):
    def __init__(
            self,
            document=None,
            feature=None,
            slce=None,
            value=None):
        self.value = value
        self.slce = slce
        self.feature = feature
        self.document = document

    def __repr__(self):
        return '''RequestContext(
    document={document},
    feature={feature},
    slce={slce},
    value={value})'''.format(**self.__dict__)

    def __str__(self):
        return self.__repr__()


class DefaultSerializer(object):
    def __init__(self, content_type):
        super(DefaultSerializer, self).__init__()
        self._content_type = content_type

    def matches(self, context):
        if context.feature is None:
            return False
        return context.feature.encoder.content_type == self._content_type

    @property
    def content_type(self):
        return self._content_type

    def serialize(self, context):
        document = context.document
        feature = context.feature
        slce = context.slce
        flo = feature(_id=document._id, decoder=Decoder(), persistence=document)
        if slce.start:
            flo.seek(slce.start)
        if slce.stop:
            value = flo.read(slce.stop - slce.start)
        else:
            value = flo.read()
        key = document.key_builder.build(document._id, feature.key)
        total = document.database.size(key)
        return TempResult(
                value,
                self.content_type,
                is_partial=slce.start is not None or slce.stop is not None,
                content_range=ContentRange.from_slice(slce, total))


class AudioSamplesSerializer(object):

    def __init__(self):
        super(AudioSamplesSerializer, self).__init__()

    def matches(self, context):
        return isinstance(context.value, AudioSamples)

    @property
    def content_type(self):
        return 'audio/ogg'

    def serialize(self, context):
        bio = BytesIO()
        samples = context.value
        with SoundFile(
                bio,
                mode='w',
                samplerate=samples.samples_per_second,
                channels=samples.channels,
                format='OGG',
                subtype='VORBIS') as sf:
            sf.write(samples)
        bio.seek(0)
        return TempResult(bio.read(), 'audio/ogg')


class OggVorbisSerializer(object):
    """
    Serializer capable of handling range requests against ogg vorbis files
    """

    def __init__(self):
        super(OggVorbisSerializer, self).__init__()

    def matches(self, context):
        if context.feature is None:
            return False
        return \
            isinstance(context.feature, OggVorbisFeature) \
            and isinstance(context.slce, TimeSlice)

    @property
    def content_type(self):
        return 'audio/ogg'

    def serialize(self, context):
        feature = context.feature
        document = context.document
        slce = context.slce
        wrapper = feature(_id=document._id, persistence=document)
        samples = wrapper[slce]
        bio = BytesIO()
        with SoundFile(
                bio,
                mode='w',
                samplerate=wrapper.samplerate,
                channels=wrapper.channels,
                format='OGG',
                subtype='VORBIS') as sf:
            sf.write(samples)
        bio.seek(0)
        content_range = ContentRange.from_timeslice(
                slce, Picoseconds(int(1e12 * wrapper.duration_seconds)))
        return TempResult(
                bio.read(),
                'audio/ogg',
                is_partial=slce != TimeSlice(),
                content_range=content_range)


def generate_image(data, is_partial=False, content_range=None):
    fig = plt.figure()
    if len(data.shape) == 1:
        plt.plot(data)
    elif len(data.shape) == 2:
        mat = plt.matshow(np.rot90(data), cmap=plt.cm.gray)
        mat.axes.get_xaxis().set_visible(False)
        mat.axes.get_yaxis().set_visible(False)
    else:
        raise ValueError('cannot handle dimensions > 2')
    bio = BytesIO()
    plt.savefig(bio, bbox_inches='tight', pad_inches=0, format='png')
    bio.seek(0)
    fig.clf()
    plt.close()
    return TempResult(
            bio.read(),
            'image/png',
            is_partial=is_partial,
            content_range=content_range)


class ConstantRateTimeSeriesSerializer(object):
    def __init__(self):
        super(ConstantRateTimeSeriesSerializer, self).__init__()

    def matches(self, context):
        return \
            isinstance(context.feature, ConstantRateTimeSeriesFeature) \
            and isinstance(context.slce, TimeSlice)

    @property
    def content_type(self):
        return 'image/png'

    def serialize(self, context):
        feature = context.feature
        document = context.document
        data = feature(_id=document._id, persistence=document)
        sliced_data = data[context.slce]
        content_range = ContentRange.from_timeslice(
                context.slce, data.end)
        return generate_image(
                sliced_data,
                is_partial=True,
                content_range=content_range)


class NumpySerializer(object):
    def __init__(self):
        super(NumpySerializer, self).__init__()

    def matches(self, context):
        if context.document is not None \
                and isinstance(context.feature, ConstantRateTimeSeriesFeature):
            return True

        return \
            isinstance(context.value, np.ndarray) \
            and len(context.value.shape) in (1, 2)

    @property
    def content_type(self):
        return 'image/png'

    def serialize(self, context):
        feature = context.feature
        document = context.document
        value = context.value
        if value is None:
            data = feature(_id=document._id, persistence=document)
        else:
            data = value
        return generate_image(data)


class AudioSliceSerializer(object):
    def __init__(
            self,
            content_type,
            visualization_feature,
            audio_feature):
        self.audio_feature = audio_feature
        self.visualization_feature = visualization_feature
        self._content_type = content_type

    @property
    def content_type(self):
        return self._content_type

    def _seconds(self, ts):
        return {
            'start': ts.start / Seconds(1),
            'duration': ts.duration / Seconds(1)
        }

    def _result(self, ts, _id):
        template = '/zounds/{_id}/{feature}'
        return {
            'audio': template.format(_id=_id, feature=self.audio_feature.key),
            'visualization': template.format(
                    _id=_id, feature=self.visualization_feature.key),
            'slice': self._seconds(ts)
        }

    def iter_results(self, context):
        raise NotImplementedError()

    def serialize(self, context):
        results = map(lambda x: self._result(*x), self.iter_results(context))
        output = {'results': results}
        return TempResult(json.dumps(output), self.content_type)


class OnsetsSerializer(AudioSliceSerializer):
    def __init__(self, visualization_feature, audio_feature):
        super(OnsetsSerializer, self).__init__(
                'application/vnd.zounds.onsets+json',
                visualization_feature,
                audio_feature)

    def matches(self, context):
        return isinstance(context.feature, TimeSliceFeature)

    def iter_results(self, context):
        for ts in context.value:
            yield ts, context.document._id


class SearchResultsSerializer(AudioSliceSerializer):
    def __init__(self, visualization_feature, audio_feature):
        super(SearchResultsSerializer, self).__init__(
                'application/vnd.zounds.searchresults+json',
                visualization_feature,
                audio_feature)

    def matches(self, context):
        return isinstance(context.value, SearchResults)

    def iter_results(self, context):
        for _id, ts in context.value:
            yield ts, _id


class ZoundsApp(object):
    def __init__(
            self,
            base_path=r'/zounds/',
            model=None,
            visualization_feature=None,
            audio_feature=None,
            globals={},
            locals={}):

        super(ZoundsApp, self).__init__()
        self.locals = locals
        self.globals = globals
        self.model = model
        self.visualization_feature = visualization_feature
        self.audio_feature = audio_feature
        self.base_path = base_path
        self.serializers = [
            AudioSamplesSerializer(),
            OggVorbisSerializer(),
            ConstantRateTimeSeriesSerializer(),
            DefaultSerializer('application/json'),
            DefaultSerializer('audio/ogg'),
            NumpySerializer(),
            OnsetsSerializer(
                    self.visualization_feature,
                    self.audio_feature),
            SearchResultsSerializer(
                    self.visualization_feature,
                    self.audio_feature)
        ]
        self.temp = {}

        path, fn = os.path.split(__file__)
        with open(os.path.join(path, 'zounds.js')) as f:
            script = f.read()

        with open(os.path.join(path, 'index.html')) as f:
            self._html_content = f.read().replace(
                    '<script src="/zounds.js"></script>',
                    '<script>{}</script>'.format(script))

    def find_serializer(self, context):
        try:
            return filter(
                    lambda x: x.matches(context), self.serializers)[0]
        except IndexError:
            raise NoMatchingSerializerException()

    def serialize(self, context):
        serializer = self.find_serializer(context)
        return serializer.serialize(context)

    def temp_handler(self):

        app = self

        class TempHandler(tornado.web.RequestHandler):

            def get(self, _id):
                try:
                    result = app.temp[_id]
                except KeyError:
                    self.set_status(httplib.NOT_FOUND)
                    self.finish()
                    return
                self.set_header('Content-Type', result.content_type)
                self.set_header('Accept-Ranges', 'bytes')
                self.write(result.data)
                self.set_status(httplib.OK)
                self.finish()

        return TempHandler

    def repl_handler(self):
        document = self.model
        globals = self.globals
        locals = self.locals
        app = self

        class ReplHandler(tornado.web.RequestHandler):

            def _extract_feature(self, statement):
                root = ast.parse(statement)
                nodes = list(ast.walk(root))
                doc = None
                feature = None
                feature_name = None

                for node in nodes:
                    if doc and feature_name:
                        break

                    if isinstance(node, ast.Name) \
                            and node.id in locals \
                            and isinstance(locals[node.id], document):
                        doc = locals[node.id]
                        continue

                    if isinstance(node, ast.Attribute) \
                            and node.attr in document.features:
                        feature_name = node.attr
                        continue

                if feature_name:
                    feature = document.features[feature_name]

                return doc, feature

            def _add_url(self, statement, output, value):
                doc, feature = self._extract_feature(statement)
                try:
                    context = RequestContext(
                            document=doc,
                            feature=feature,
                            value=value,
                            slce=slice(None))
                    result = app.serialize(context)
                    temp_id = uuid.uuid4().hex
                    app.temp[temp_id] = result
                    output['url'] = '/zounds/temp/{temp_id}'.format(
                            temp_id=temp_id)
                    output['contentType'] = result.content_type
                except NoMatchingSerializerException:
                    pass

            def post(self):
                statement = self.request.body
                self.set_header('Content-Type', 'application/json')
                output = dict()

                try:
                    try:
                        value = eval(statement, globals, locals)
                        output['result'] = str(value)
                        self._add_url(statement, output, value)
                    except SyntaxError:
                        orig_stdout = sys.stdout
                        sys.stdout = sio = StringIO()
                        exec (statement, globals, locals)
                        sio.seek(0)
                        output['result'] = sio.read()
                        sys.stdout = orig_stdout
                    self.set_status(httplib.OK)
                except:
                    output['error'] = traceback.format_exc()
                    self.set_status(httplib.BAD_REQUEST)

                self.write(json.dumps(output))
                self.finish()

        return ReplHandler

    def feature_handler(self):

        document = self.model
        app = self

        class FeatureHandler(tornado.web.RequestHandler):

            def get(self, _id, feature):
                doc = document(_id)
                feature = document.features[feature]
                try:
                    slce = RangeRequest(self.request.headers['Range']).range()
                except KeyError:
                    slce = slice(None)
                context = RequestContext(
                        document=doc, feature=feature, slce=slce)
                try:
                    result = app.serialize(context)
                except UnsatisfiableRangeRequestException:
                    self.set_status(httplib.REQUESTED_RANGE_NOT_SATISFIABLE)
                    self.finish()
                self.set_header('Content-Type', result.content_type)
                self.set_header('Accept-Ranges', 'bytes')
                self.write(result.data)
                self.set_header('ETag', self.compute_etag())
                self.set_status(
                        httplib.PARTIAL_CONTENT if result.is_partial
                        else httplib.OK)
                if result.content_range:
                    self.set_header('Content-Range', str(result.content_range))
                self.finish()

        return FeatureHandler

    def ui_handler(self):
        app = self

        class UIHandler(tornado.web.RequestHandler):
            def get(self):
                self.set_header('Content-Type', 'text-html')
                self.write(app._html_content)
                self.set_status(httplib.OK)
                self.finish()

        return UIHandler

    def build_app(self):
        return tornado.web.Application([
            (r'/', self.ui_handler()),
            (r'/zounds/temp/(.+?)/?', self.temp_handler()),
            (r'/zounds/repl/?', self.repl_handler()),
            (r'/zounds/(.+?)/(.+?)/?', self.feature_handler())
        ])

    def start(self, port=8888):
        app = self.build_app()
        app.listen(port)
        tornado.ioloop.IOLoop.current().start()



import unittest2
import featureflow
from random_samples import ReservoirSampler
from preprocess import \
    UnitNorm, MeanStdNormalization, PreprocessingPipeline, Pipeline
from learn import LinearRbm
import numpy as np


class Iterator(featureflow.Node):
    def __init__(self, needs=None):
        super(Iterator, self).__init__(needs=needs)

    def _process(self, data):
        for d in data:
            yield d


class RbmTests(unittest2.TestCase):
    def test_can_retrieve_rbm_pipeline(self):
        class Settings(featureflow.PersistenceSettings):
            _id = 'rbm'
            id_provider = featureflow.StaticIdProvider(_id)
            key_builder = featureflow.StringDelimitedKeyBuilder()
            database = featureflow.InMemoryDatabase(key_builder=key_builder)

        class Learned(featureflow.BaseModel, Settings):
            iterator = featureflow.Feature(
                    Iterator,
                    store=False)

            shuffle = featureflow.NumpyFeature(
                    ReservoirSampler,
                    nsamples=1000,
                    needs=iterator,
                    store=True)

            unitnorm = featureflow.PickleFeature(
                    UnitNorm,
                    needs=shuffle,
                    store=False)

            meanstd = featureflow.PickleFeature(
                    MeanStdNormalization,
                    needs=unitnorm,
                    store=False)

            rbm = featureflow.PickleFeature(
                    LinearRbm,
                    hdim=64,
                    epochs=5,
                    needs=meanstd,
                    store=False)

            pipeline = featureflow.PickleFeature(
                    PreprocessingPipeline,
                    needs=(unitnorm, meanstd, rbm),
                    store=True)

        def data():
            for i in xrange(100):
                yield np.random.random_sample((np.random.randint(10, 100), 3))

        Learned.process(iterator=data())
        self.assertIsInstance(Learned().pipeline, Pipeline)



import unittest2
import numpy as np
from featureflow import BaseModel, Node, PersistenceSettings
from basic import Merge
from zounds.timeseries import \
    ConstantRateTimeSeries, ConstantRateTimeSeriesFeature, Milliseconds


class MergeTester(Node):
    def __init__(
            self,
            total_frames=100,
            increments_of=30,
            features=10,
            needs=None):
        super(MergeTester, self).__init__(needs=needs)
        self.features = features
        self.increments_of = increments_of
        self.total_frames = total_frames

    def _process(self, data):
        for i in xrange(0, self.total_frames, self.increments_of):
            size = min(self.increments_of, self.total_frames - i)
            yield ConstantRateTimeSeries(
                    np.zeros((size, self.features)),
                    frequency=Milliseconds(500))


class MergeTests(unittest2.TestCase):
    def test_raises_if_single_source(self):
        class Document(BaseModel, PersistenceSettings):
            source = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    store=True)
            merged = ConstantRateTimeSeriesFeature(
                    Merge,
                    needs=source,
                    store=True)

        self.assertRaises(ValueError, lambda: Document.process(source=''))

    def test_raises_if_single_element_iterable(self):
        class Document(BaseModel, PersistenceSettings):
            source = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    store=True)
            merged = ConstantRateTimeSeriesFeature(
                    Merge,
                    needs=[source],
                    store=True)

        self.assertRaises(ValueError, lambda: Document.process(source=''))

    def test_can_combine_two_sources_at_same_rate(self):
        class Document(BaseModel, PersistenceSettings):
            source1 = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    total_frames=200,
                    store=True)
            source2 = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    total_frames=200,
                    store=True)
            merged = ConstantRateTimeSeriesFeature(
                    Merge,
                    needs=[source1, source2],
                    store=True)

        _id = Document.process(source1='', source2='')
        doc = Document(_id)
        self.assertEqual((200, 20), doc.merged.shape)

    def test_can_combine_three_sources_at_same_rate(self):
        class Document(BaseModel, PersistenceSettings):
            source1 = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    total_frames=200,
                    store=True)
            source2 = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    total_frames=200,
                    store=True)
            source3 = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    total_frames=200,
                    store=True)
            merged = ConstantRateTimeSeriesFeature(
                    Merge,
                    needs=(source1, source2, source3),
                    store=True)

        _id = Document.process(source1='', source2='', source3='')
        doc = Document(_id)
        self.assertEqual((200, 30), doc.merged.shape)

    def test_can_combine_two_sources_at_different_rates(self):
        class Document(BaseModel, PersistenceSettings):
            source1 = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    total_frames=200,
                    increments_of=30,
                    store=True)
            source2 = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    total_frames=200,
                    increments_of=40,
                    store=True)
            merged = ConstantRateTimeSeriesFeature(
                    Merge,
                    needs=(source1, source2),
                    store=True)

        _id = Document.process(source1='', source2='')
        doc = Document(_id)
        self.assertEqual((200, 20), doc.merged.shape)

    def test_can_combine_three_sources_at_different_rates(self):
        class Document(BaseModel, PersistenceSettings):
            source1 = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    total_frames=200,
                    increments_of=12,
                    store=True)
            source2 = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    total_frames=200,
                    increments_of=17,
                    store=True)
            source3 = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    total_frames=200,
                    increments_of=32,
                    store=True)
            merged = ConstantRateTimeSeriesFeature(
                    Merge,
                    needs=(source1, source2, source3),
                    store=True)

        _id = Document.process(source1='', source2='', source3='')
        doc = Document(_id)
        self.assertEqual((200, 30), doc.merged.shape)

    def test_shortest_of_two_sources(self):
        class Document(BaseModel, PersistenceSettings):
            source1 = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    total_frames=190,
                    increments_of=30,
                    store=True)
            source2 = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    total_frames=200,
                    increments_of=40,
                    store=True)
            merged = ConstantRateTimeSeriesFeature(
                    Merge,
                    needs=[source1, source2],
                    store=True)

        _id = Document.process(source1='', source2='')
        doc = Document(_id)
        self.assertEqual((190, 20), doc.merged.shape)

    def test_shortest_of_three_sources(self):
        class Document(BaseModel, PersistenceSettings):
            source1 = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    total_frames=200,
                    increments_of=12,
                    store=True)
            source2 = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    total_frames=185,
                    increments_of=17,
                    store=True)
            source3 = ConstantRateTimeSeriesFeature(
                    MergeTester,
                    total_frames=50,
                    increments_of=32,
                    store=True)
            merged = ConstantRateTimeSeriesFeature(
                    Merge,
                    needs=[source1, source2, source3],
                    store=True)

        _id = Document.process(source1='', source2='', source3='')
        doc = Document(_id)
        self.assertEqual((50, 30), doc.merged.shape)



from featureflow import Node, NotEnoughData
import marshal
import types
from collections import OrderedDict


class Op(object):
    def __init__(self, func, **kwargs):
        super(Op, self).__init__()
        self._kwargs = kwargs
        self._func = marshal.dumps(func.func_code)

    def __getattr__(self, key):
        if key == '_kwargs':
            raise AttributeError()

        try:
            return self._kwargs[key]
        except KeyError:
            raise AttributeError(key)

    def __call__(self, arg, **kwargs):
        code = marshal.loads(self._func)
        f = types.FunctionType(
                code,
                globals(),
                'preprocess')
        kwargs.update(self._kwargs)
        return f(arg, **kwargs)


class PreprocessResult(object):
    def __init__(self, data, op, inversion_data=None, inverse=None, name=None):
        super(PreprocessResult, self).__init__()
        self.name = name
        self.inverse = inverse
        self.inversion_data = inversion_data
        self.data = data
        self.op = op


class Preprocessor(Node):
    def __init__(self, needs=None):
        super(Preprocessor, self).__init__(needs=needs)

    def _forward_func(self):
        """
        Return a function that represents this processor's
        forward transform
        """

        def x(data, example_arg=None, another_one=None):
            return data

        return x

    def transform(self, **kwargs):
        return Op(self._forward_func(), **kwargs)

    def _inversion_data(self):
        """
        Return a function that computes any data needed for
        an inverse transform, if one is possible.  Otherwise,
        raise NotImplemented
        """

        def x(data, **kwargs):
            return kwargs

        return x

    def inversion_data(self, **kwargs):
        return Op(self._inversion_data(), **kwargs)

    def _backward_func(self):
        """
        Return a function that computes this processor's
        inverse transform, if one is possible.  Otherwise,
        raise NotImplemented
        """

        def x(data, **inversion_args):
            return data

        return x

    def inverse_transform(self):
        return Op(self._backward_func())

    def _extract_data(self, data):
        if isinstance(data, PreprocessResult):
            return data.data
        return data


class UnitNorm(Preprocessor):
    def __init__(self, needs=None):
        super(UnitNorm, self).__init__(needs=needs)

    def _forward_func(self):
        def x(d):
            from zounds.nputil import safe_unit_norm
            return safe_unit_norm(d.reshape(d.shape[0], -1))

        return x

    def _inversion_data(self):
        def x(d):
            import numpy as np
            return dict(norm=np.linalg.norm(d, axis=1))

        return x

    def _backward_func(self):
        def x(d, norm=None):
            return d * norm[:, None]

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform()
        inv_data = self.inversion_data()
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
                data, op, inversion_data=inv_data, inverse=inv, name='UnitNorm')


class Log(Preprocessor):
    def __init__(self, needs=None):
        super(Log, self).__init__(needs=needs)

    def _forward_func(self):
        def x(d):
            import numpy as np
            m = np.min(d)
            if m > 0:
                return np.log(m)
            pos = d + -np.min(d) + 1
            return np.log(pos)

        return x

    def _inversion_data(self):
        def x(d):
            import numpy as np
            return dict(min=np.min(d))

        return x

    def _backward_func(self):
        def x(d, min=None):
            import numpy as np
            if min > 0:
                return np.exp(d)
            return np.exp(d) - (-min) - 1

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform()
        inv_data = self.inversion_data()
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
                data, op, inversion_data=inv_data, inverse=inv, name='Log')


class MeanStdNormalization(Preprocessor):
    def __init__(self, needs=None):
        super(MeanStdNormalization, self).__init__(needs=needs)

    def _forward_func(self):
        def x(d, mean=None, std=None):
            return (d - mean) / std

        return x

    def _backward_func(self):
        def x(d, mean=None, std=None):
            arr = d.copy()
            arr *= std
            arr += mean
            return arr

        return x

    def _process(self, data):
        data = self._extract_data(data)
        mean = data.mean(axis=0)
        std = data.std(axis=0)
        op = self.transform(mean=mean, std=std)
        inv_data = self.inversion_data(mean=mean, std=std)
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
                data, op, inversion_data=inv_data, inverse=inv, name='MeanStd')


class Binarize(Preprocessor):
    def __init__(self, threshold=0.5, needs=None):
        super(Binarize, self).__init__(needs=needs)
        self.threshold = threshold

    def _forward_func(self):
        def x(d, threshold=None):
            import numpy as np
            data = np.zeros(d.shape, dtype=np.uint8)
            data[np.where(d > threshold)] = 1
            return data

        return x

    def _backward_func(self):
        def x(d):
            raise NotImplementedError()

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform(threshold=self.threshold)
        inv_data = self.inversion_data()
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
                data, op, inversion_data=inv_data, inverse=inv, name='Binarize')


class Pipeline(object):
    def __init__(self, preprocess_results):
        self.processors = preprocess_results

    def transform(self, data):
        inversion_data = []
        for p in self.processors:
            inversion_data.append(p.inversion_data(data))
            data = p.op(data)
        return PipelineResult(data, self.processors, inversion_data)


class PipelineResult(object):
    def __init__(self, data, processors, inversion_data):
        super(PipelineResult, self).__init__()
        self.processors = processors[::-1]
        self.inversion_data = inversion_data[::-1]
        self.data = data

    def inverse_transform(self):
        data = self.data
        for inv_data, p in zip(self.inversion_data, self.processors):
            data = p.inverse(data, **inv_data)
        return data


class PreprocessingPipeline(Node):
    def __init__(self, needs=None):
        super(PreprocessingPipeline, self).__init__(needs=needs)
        self._pipeline = OrderedDict((id(n), None) for n in needs)

    def _enqueue(self, data, pusher):
        self._pipeline[id(pusher)] = data

    def _dequeue(self):
        # TODO: Make this an aggregator
        if not self._finalized or not all(self._pipeline.itervalues()):
            raise NotEnoughData()

        return Pipeline(self._pipeline.values())



from featureflow import Node, NotEnoughData
import numpy as np
from collections import OrderedDict
from zounds.timeseries import ConstantRateTimeSeries


class Merge(Node):
    """
    Combine two or more sources into a single feature
    """
    def __init__(self, needs=None):
        super(Merge, self).__init__(needs=needs)
        exc_msg = 'you must supply two or more dependencies'
        try:
            needs = list(needs)
        except TypeError:
            raise ValueError(exc_msg)

        if len(needs) < 2:
            raise ValueError(exc_msg)

        self._cache = OrderedDict((id(n), None) for n in needs)

    def _enqueue(self, data, pusher):
        key = id(pusher)
        print self._cache[key]
        if self._cache[key] is None or self._cache[key].size == 0:
            self._cache[key] = data
        else:
            self._cache[key] = self._cache[key].concatenate(data)

    def _dequeue(self):
        if any(v is None or len(v) == 0 for v in self._cache.itervalues()):
            raise NotEnoughData()
        shortest = min(len(v) for v in self._cache.itervalues())
        output = OrderedDict(
            (k, v[:shortest]) for k, v in self._cache.iteritems())
        self._cache = OrderedDict(
            (k, v[shortest:]) for k, v in self._cache.iteritems())
        return output

    def _process(self, data):
        yield ConstantRateTimeSeries.concat(data.values(), axis=1)


class Slice(Node):
    def __init__(self, sl=None, needs=None):
        super(Slice, self).__init__(needs=needs)
        self._sl = sl

    def _process(self, data):
        yield data[:, self._sl]


class Sum(Node):
    def __init__(self, axis=0, needs=None):
        super(Sum, self).__init__(needs=needs)
        self._axis = axis

    def _process(self, data):
        # TODO: This should be generalized.  Sum will have this same problem
        try:
            data = np.sum(data, axis=self._axis)
        except ValueError:
            print 'ERROR'
            data = data
        if data.shape[0]:
            yield data


class Max(Node):
    def __init__(self, axis=0, needs=None):
        super(Max, self).__init__(needs=needs)
        self._axis = axis

    def _process(self, data):
        # TODO: This should be generalized.  Sum will have this same problem
        try:
            data = np.max(data, axis=self._axis)
        except ValueError:
            print 'ERROR'
            data = data
        if data.shape[0]:
            yield data




import unittest2
from api import RangeRequest, RangeUnitUnsupportedException, ContentRange
from zounds.timeseries import TimeSlice, Seconds, Picoseconds, Milliseconds


class ContentRangeTests(unittest2.TestCase):

    def test_open_content_range(self):
        self.assertEqual(
            'bytes 10-100/100',
            str(ContentRange('bytes', 10, 100)))

    def test_closed_content_range(self):
        self.assertEqual(
            'seconds 10-90/100',
            str(ContentRange('seconds', 10, 100, stop=90)))

    def test_from_timeslce_full_slice(self):
        ts = TimeSlice()
        self.assertEqual(
            'seconds 0.0-100.0/100.0',
            str(ContentRange.from_timeslice(ts, Seconds(100))))

    def test_from_timeslice_open_ended(self):
        ts = TimeSlice(start=Picoseconds(int(1e12)) * 2.5)
        self.assertEqual(
            'seconds 2.5-100.0/100.0',
            str(ContentRange.from_timeslice(ts, Seconds(100))))

    def test_from_timeslice_closed(self):
        ts = TimeSlice(
            start=Picoseconds(int(1e12)) * 2.5,
            duration=Milliseconds(2000))
        self.assertEqual(
            'seconds 2.5-4.5/100.0',
            str(ContentRange.from_timeslice(ts, Seconds(100))))


class RangeRequestTests(unittest2.TestCase):

    def test_raises_for_unsupported_unit(self):
        rr = RangeRequest('hours=1-2')
        self.assertRaises(RangeUnitUnsupportedException, lambda: rr.range())

    def test_can_get_open_ended_byte_slice(self):
        rr = RangeRequest('bytes=10-')
        sl = rr.range()
        self.assertIsInstance(sl, slice)
        self.assertEqual((10, None, None), (sl.start, sl.stop, sl.step))

    def test_can_get_closed_byte_slice(self):
        rr = RangeRequest('bytes=10-100')
        sl = rr.range()
        self.assertIsInstance(sl, slice)
        self.assertEqual((10, 100, None), (sl.start, sl.stop, sl.step))

    def test_can_get_open_ended_time_slice(self):
        rr = RangeRequest('seconds=0-')
        sl = rr.range()
        self.assertIsInstance(sl, TimeSlice)
        self.assertEqual(TimeSlice(start=Seconds(0)), sl)

    def test_can_get_closed_time_slice(self):
        rr = RangeRequest('seconds=10.5-100.5')
        sl = rr.range()
        self.assertIsInstance(sl, TimeSlice)
        expected_start = Picoseconds(int(10.5 * 1e12))
        expected_duration = Picoseconds(int(90 * 1e12))
        self.assertEqual(
            TimeSlice(start=expected_start, duration=expected_duration), sl)



from featureflow import Node, Feature, DatabaseIterator, BaseModel, NotEnoughData
from featureflow.nmpy import NumpyFeature
from zounds.spectral import NDSlidingWindow
import numpy as np


class ReservoirSampler(Node):
    '''
    Use reservoir sampling (http://en.wikipedia.org/wiki/Reservoir_sampling) to
    draw a fixed-size set of random samples from a stream of unknown size.
    
    This is useful when the samples can fit into memory, but the stream cannot.
    '''

    def __init__(self, needs=None, nsamples=None):
        super(ReservoirSampler, self).__init__(needs=needs)
        self._nsamples = nsamples
        self._r = None
        self._index = 0

    # TODO: What happens if we have filled up all the sample slots and we run
    # out of data?
    def _enqueue(self, data, pusher):
        if self._r is None:
            shape = (self._nsamples,) + data.shape[1:]
            self._r = np.zeros(shape, dtype=data.dtype)

        diff = 0
        if self._index < self._nsamples:
            diff = self._nsamples - self._index
            available = len(data[:diff])
            self._r[self._index: self._index + available] = data[:diff]
            self._index += available

        remaining = len(data[diff:])
        if not remaining: return
        indices = np.random.random_integers(0, self._index, size=remaining)
        indices = indices[indices < self._nsamples]
        self._r[indices] = data[diff:][range(len(indices))]
        self._index += remaining

    def _dequeue(self):
        if not self._finalized:
            raise NotEnoughData()

        if self._index <= self._nsamples:
            arr = self._r[:self._index]
            np.random.shuffle(arr)
            return arr

        return self._r


class RequireSignalToNoiseRatio(Node):
    def __init__(self, max_snr=None, needs=None):
        super(RequireSignalToNoiseRatio, self).__init__(needs=needs)
        self._max_snr = max_snr

    def _process(self, data):
        axes = tuple(range(len(data.shape))[1:])
        a = np.abs(data)
        snr = a.mean(axis=axes) / a.std(axis=axes)
        filtered = data[snr < self._max_snr]
        print len(data), len(filtered)
        yield filtered


def random_samples(feature_func, nsamples, store_shuffled=True):
    '''
    Return a base class that samples randomly from examples of the data returned
    by feature_func
    '''

    class RandomSamples(BaseModel):
        docs = Feature( \
                DatabaseIterator,
                func=feature_func,
                store=False)

        patches = NumpyFeature( \
                ReservoirSampler,
                nsamples=nsamples,
                needs=docs,
                store=store_shuffled)

    return RandomSamples


def random_patches( \
        feature_func,
        windowsize,
        nsamples,
        stepsize=None,
        store_shuffled=True):
    '''
    Return a base class that samples random, n-dimensional patches from a 
    feature, e.g., 10x10 patches from spectrograms
    '''

    if stepsize is None:
        stepsize = (1,) * len(windowsize)

    class RandomPatches(BaseModel):
        docs = Feature( \
                DatabaseIterator,
                func=feature_func,
                store=False)

        windowed = NumpyFeature( \
                NDSlidingWindow,
                windowsize=windowsize,
                stepsize=stepsize,
                needs=docs,
                store=False)

        patches = NumpyFeature( \
                ReservoirSampler,
                nsamples=nsamples,
                needs=windowed,
                store=store_shuffled)

    return RandomPatches



from featureflow import BaseModel, JSONFeature, ByteStream, ByteStreamFeature
from zounds.soundfile import \
    MetaData, AudioMetaDataEncoder, OggVorbis, OggVorbisFeature, AudioStream, \
    Resampler
from zounds.timeseries import ConstantRateTimeSeriesFeature, SR44100, HalfLapped
from zounds.spectral import \
    SlidingWindow, OggVorbisWindowingFunc, FFT, BarkBands


def audio_graph(
        chunksize_bytes=2 * 44100 * 30 * 2,
        resample_to=SR44100(),
        freesound_api_key=None,
        store_fft=False):
    """
    Produce a base class suitable as a starting point for many audio processing
    pipelines.  This class resamples all audio to a common sampling rate, and
    produces a bark band spectrogram from overlapping short-time fourier
    transform frames.  It also compresses the audio into ogg vorbis format for
    compact storage.
    """

    class AudioGraph(BaseModel):
        meta = JSONFeature(
                MetaData,
                freesound_api_key=freesound_api_key,
                store=True,
                encoder=AudioMetaDataEncoder)

        raw = ByteStreamFeature(
                ByteStream,
                chunksize=chunksize_bytes,
                needs=meta,
                store=False)

        ogg = OggVorbisFeature(
                OggVorbis,
                needs=raw,
                store=True)

        pcm = ConstantRateTimeSeriesFeature(
                AudioStream,
                needs=raw,
                store=False)

        resampled = ConstantRateTimeSeriesFeature(
                Resampler,
                needs=pcm,
                samplerate=resample_to)

        windowed = ConstantRateTimeSeriesFeature(
                SlidingWindow,
                needs=resampled,
                wscheme=HalfLapped(),
                wfunc=OggVorbisWindowingFunc(),
                store=False)

        fft = ConstantRateTimeSeriesFeature(
                FFT,
                needs=windowed,
                store=store_fft)

        short_windowed = ConstantRateTimeSeriesFeature(
                SlidingWindow,
                needs=resampled,
                wscheme=HalfLapped(512, 256),
                wfunc=OggVorbisWindowingFunc(),
                store=False)

        short_fft = ConstantRateTimeSeriesFeature(
                FFT,
                needs=short_windowed,
                store=store_fft)

        bark = ConstantRateTimeSeriesFeature(
                BarkBands,
                needs=fft,
                samplerate=resample_to,
                store=True)

    return AudioGraph



import unittest2
import numpy as np
from duration import Seconds
from samplerate import SR44100, SR11025, SampleRate
from audiosamples import AudioSamples


class AudioSamplesTest(unittest2.TestCase):
    def test_raises_if_not_audio_samplerate(self):
        arr = np.zeros(44100 * 2.5)
        one = Seconds(1)
        self.assertRaises(
                TypeError, lambda: AudioSamples(arr, SampleRate(one, one)))

    def test_can_create_instance(self):
        arr = np.zeros(44100 * 2.5)
        instance = AudioSamples(arr, SR44100())
        self.assertIsInstance(instance, AudioSamples)
        length_seconds = instance.end / Seconds(1)
        self.assertAlmostEqual(2.5, length_seconds, places=6)

    def test_channels_returns_one_for_one_dimensional_array(self):
        arr = np.zeros(44100 * 2.5)
        instance = AudioSamples(arr, SR44100())
        self.assertEqual(1, instance.channels)

    def test_channels_returns_two_for_two_dimensional_array(self):
        arr = np.zeros(44100 * 2.5)
        arr = np.column_stack((arr, arr))
        instance = AudioSamples(arr, SR44100())
        self.assertEqual(2, instance.channels)

    def test_samplerate_returns_correct_value(self):
        arr = np.zeros(44100 * 2.5)
        instance = AudioSamples(arr, SR44100())
        self.assertIsInstance(instance.samplerate, SR44100)

    def test_can_sum_to_mono(self):
        arr = np.zeros(44100 * 2.5)
        arr = np.column_stack((arr, arr))
        instance = AudioSamples(arr, SR44100())
        mono = instance.mono
        self.assertEqual(1, mono.channels)
        self.assertIsInstance(mono.samplerate, SR44100)

    def test_class_concat_returns_audio_samples(self):
        s1 = AudioSamples(np.zeros(44100 * 2), SR44100())
        s2 = AudioSamples(np.zeros(44100), SR44100())
        s3 = AudioSamples.concat([s1, s2])
        self.assertIsInstance(s3, AudioSamples)
        self.assertEqual(44100 * 3, len(s3))

    def test_instance_concat_returns_audio_samples(self):
        s1 = AudioSamples(np.zeros(44100 * 2), SR44100())
        s2 = AudioSamples(np.zeros(44100), SR44100())
        s3 = s1.concat([s1, s2])
        self.assertIsInstance(s3, AudioSamples)
        self.assertEqual(44100 * 3, len(s3))

    def test_concat_raises_for_different_sample_rates(self):
        s1 = AudioSamples(np.zeros(44100 * 2), SR44100())
        s2 = AudioSamples(np.zeros(44100), SR11025())
        self.assertRaises(ValueError, lambda: AudioSamples.concat([s1, s2]))






import unittest2
import featureflow
import numpy as np
from preprocess import \
    UnitNorm, MeanStdNormalization, Binarize, PreprocessingPipeline


class TestPipeline(unittest2.TestCase):
    def test_cannot_invert_pipeline_if_any_steps_are_missing(self):
        class Settings(featureflow.PersistenceSettings):
            id_provider = featureflow.UuidProvider()
            key_builder = featureflow.StringDelimitedKeyBuilder()
            database = featureflow.InMemoryDatabase(key_builder=key_builder)

        class Model(featureflow.BaseModel, Settings):
            unitnorm = featureflow.PickleFeature(
                    UnitNorm,
                    store=False)

            binary = featureflow.PickleFeature(
                    Binarize,
                    needs=unitnorm,
                    store=False)

            pipeline = featureflow.PickleFeature(
                    PreprocessingPipeline,
                    needs=(unitnorm, binary),
                    store=True)

        data = np.random.random_sample((1000, 4))
        _id = Model.process(unitnorm=data)
        example = np.random.random_sample((10, 4))
        model = Model(_id)
        transformed = model.pipeline.transform(example)
        self.assertRaises(
                NotImplementedError, lambda: transformed.inverse_transform())

    def test_can_invert_pipeline(self):
        class Settings(featureflow.PersistenceSettings):
            id_provider = featureflow.UuidProvider()
            key_builder = featureflow.StringDelimitedKeyBuilder()
            database = featureflow.InMemoryDatabase(key_builder=key_builder)

        class Model(featureflow.BaseModel, Settings):
            unitnorm = featureflow.PickleFeature(
                    UnitNorm,
                    store=False)

            meanstd = featureflow.PickleFeature(
                    MeanStdNormalization,
                    needs=unitnorm,
                    store=False)

            pipeline = featureflow.PickleFeature(
                    PreprocessingPipeline,
                    needs=(unitnorm, meanstd),
                    store=True)

        data = np.random.random_sample((1000, 4))
        _id = Model.process(unitnorm=data)
        example = np.random.random_sample((10, 4))
        model = Model(_id)
        transformed = model.pipeline.transform(example)
        reconstructed = transformed.inverse_transform()
        diff = np.abs(example - reconstructed)
        self.assertTrue(np.all(diff < .00001))



from setuptools import setup
import re

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    long_description = open('README.md').read()

with open('zounds/__init__.py', 'r') as fd:
    version = re.search(
            r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
            fd.read(),
            re.MULTILINE).group(1)

download_url = 'https://github.com/jvinyard/zounds/tarball/{version}'\
    .format(**locals())

setup(
        name='zounds',
        version=version,
        url='http://www.johnvinyard.com',
        author='John Vinyard',
        author_email='john.vinyard@gmail.com',
        long_description=long_description,
        download_url=download_url,
        packages=[
            'zounds',
            'zounds.basic',
            'zounds.index',
            'zounds.learn',
            'zounds.nputil',
            'zounds.segment',
            'zounds.soundfile',
            'zounds.spectral',
            'zounds.synthesize',
            'zounds.timeseries',
            'zounds.ui'
        ],
        install_requires=[
            'featureflow',
            'nose',
            'unittest2',
            'requests',
            'tornado',
            'pysoundfile',
            'cython',
            'matplotlib'
        ],
        package_data={
            'nputil': ['*.pyx', '*.pyxbld'],
            'ui': ['*.html', '*.js']
        },
        include_package_data=True
)



import numpy as np


class Hours(np.timedelta64):
    def __new__(cls, hours):
        return np.timedelta64(hours, 'h')


class Minutes(np.timedelta64):
    def __new__(cls, minutes):
        return np.timedelta64(minutes, 'm')


class Seconds(np.timedelta64):
    def __new__(cls, seconds):
        return np.timedelta64(seconds, 's')


class Milliseconds(np.timedelta64):
    def __new__(cls, milliseconds):
        return np.timedelta64(milliseconds, 'ms')


class Microseconds(np.timedelta64):
    def __new__(cls, microseconds):
        return np.timedelta64(microseconds, 'us')


class Nanoseconds(np.timedelta64):
    def __new__(cls, nanoseconds):
        return np.timedelta64(nanoseconds, 'ns')


class Picoseconds(np.timedelta64):
    def __new__(cls, picoseconds):
        return np.timedelta64(picoseconds, 'ps')



from featureflow import NumpyMetaData, Node, Feature, Decoder
import numpy as np
from duration import Picoseconds
from samplerate import SampleRate


class TimeSlice(object):
    def __init__(self, duration=None, start=None):
        super(TimeSlice, self).__init__()

        if duration is not None and not isinstance(duration, np.timedelta64):
            raise ValueError('duration must be of type {t} but was {t2}'.format( \
                    t=np.timedelta64, t2=duration.__class__))

        if start is not None and not isinstance(start, np.timedelta64):
            raise ValueError('start must be of type {t} but was {t2}'.format( \
                    t=np.timedelta64, t2=start.__class__))

        self.duration = duration
        self.start = start or np.timedelta64(0, 's')

    def __add__(self, other):
        return TimeSlice(self.duration, start=self.start + other)

    def __radd__(self, other):
        return self.__add__(other)

    @property
    def end(self):
        return self.start + self.duration

    def __and__(self, other):
        delta = max( \
                np.timedelta64(0, 's'),
                min(self.end, other.end) - max(self.start, other.start))
        return TimeSlice(delta)

    def __contains__(self, other):
        if isinstance(other, np.timedelta64):
            return other > self.start and other < self.end
        if isinstance(other, TimeSlice):
            return other.start > self.start and other.end < self.end
        raise ValueError

    def __eq__(self, other):
        return self.start == other.start and self.duration == other.duration

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '{cls}(duration = {duration}, start = {start})'.format( \
                cls=self.__class__.__name__,
                duration=str(self.duration),
                start=str(self.start))

    def __str__(self):
        return self.__repr__()


class ConstantRateTimeSeriesMetadata(NumpyMetaData):
    def __init__( \
            self,
            dtype=None,
            shape=None,
            frequency=None,
            duration=None):
        super(ConstantRateTimeSeriesMetadata, self).__init__( \
                dtype=dtype, shape=shape)
        self.frequency = self._decode_timedelta(frequency)
        self.duration = self._decode_timedelta(duration)

    @staticmethod
    def from_timeseries(timeseries):
        return ConstantRateTimeSeriesMetadata( \
                dtype=timeseries.dtype,
                shape=timeseries.shape[1:],
                frequency=timeseries.frequency,
                duration=timeseries.duration)

    def _encode_timedelta(self, td):
        return (td.astype(np.uint64).tostring(), str(td.dtype)[-3:-1])

    def _decode_timedelta(self, t):
        if isinstance(t, np.timedelta64):
            return t

        v = np.fromstring(t[0], dtype=np.uint64)[0]
        s = t[1]
        return np.timedelta64(long(v), s)

    def __repr__(self):
        return repr((
            str(np.dtype(self.dtype)),
            self.shape,
            self._encode_timedelta(self.frequency),
            self._encode_timedelta(self.duration)
        ))


class BaseConstantRateTimeSeriesEncoder(Node):
    content_type = 'application/octet-stream'

    def __init__(self, needs=None):
        super(BaseConstantRateTimeSeriesEncoder, self).__init__(needs=needs)
        self.metadata = None

    def _prepare_data(self, data):
        raise NotImplementedError()

    def _process(self, data):
        data = self._prepare_data(data)
        if not self.metadata:
            self.metadata = ConstantRateTimeSeriesMetadata \
                .from_timeseries(data)
            packed = self.metadata.pack()
            yield packed

        encoded = data.tostring()
        yield encoded


class ConstantRateTimeSeriesEncoder(BaseConstantRateTimeSeriesEncoder):
    def _prepare_data(self, data):
        return data


class PackedConstantRateTimeSeriesEncoder(BaseConstantRateTimeSeriesEncoder):
    def __init__(self, needs=None, axis=1):
        super(PackedConstantRateTimeSeriesEncoder, self).__init__(needs=needs)
        self.axis = axis

    def _prepare_data(self, data):
        packedbits = np.packbits(data.astype(np.uint8), axis=self.axis)

        return ConstantRateTimeSeries( \
                packedbits,
                frequency=data.frequency,
                duration=data.duration)


def _np_from_buffer(b, shape, dtype, freq, duration):
    f = np.frombuffer if len(b) else np.fromstring
    shape = tuple(int(x) for x in shape)
    f = f(b, dtype=dtype).reshape(shape)
    return ConstantRateTimeSeries(f, freq, duration)


class GreedyConstantRateTimeSeriesDecoder(Decoder):
    def __init__(self):
        super(GreedyConstantRateTimeSeriesDecoder, self).__init__()

    def __call__(self, flo):
        metadata, bytes_read = ConstantRateTimeSeriesMetadata.unpack(flo)

        leftovers = flo.read()
        leftover_bytes = len(leftovers)
        first_dim = leftover_bytes / metadata.totalsize
        dim = (first_dim,) + metadata.shape
        out = _np_from_buffer( \
                leftovers,
                dim,
                metadata.dtype,
                metadata.frequency,
                metadata.duration)
        return out

    def __iter__(self, flo):
        yield self(flo)


class ConstantRateTimeSeriesFeature(Feature):
    def __init__( \
            self,
            extractor,
            needs=None,
            store=False,
            key=None,
            encoder=ConstantRateTimeSeriesEncoder,
            decoder=GreedyConstantRateTimeSeriesDecoder(),
            **extractor_args):
        super(ConstantRateTimeSeriesFeature, self).__init__( \
                extractor,
                needs=needs,
                store=store,
                encoder=encoder,
                decoder=decoder,
                key=key,
                **extractor_args)


class ConstantRateTimeSeries(np.ndarray):
    """
    A TimeSeries implementation with samples of a constant duration and
    frequency.
    """
    __array_priority__ = 10.0

    def __new__(cls, input_array, frequency, duration=None):
        if not isinstance(frequency, np.timedelta64):
            raise ValueError('duration must be of type {t} but was {t2}'.format( \
                    t=np.timedelta64, t2=frequency.__class__))

        if duration is not None and not isinstance(duration, np.timedelta64):
            raise ValueError('start must be of type {t} but was {t2}'.format(
                    t=np.timedelta64, t2=duration.__class__))

        obj = np.asarray(input_array).view(cls)
        obj.frequency = frequency
        obj.duration = duration or frequency
        return obj

    @classmethod
    def from_example(cls, arr, example):
        return cls(arr, frequency=example.frequency, duration=example.duration)

    def concatenate(self, other):
        if self.frequency == other.frequency and self.duration == other.duration:
            return self.from_example(np.concatenate([self, other]), self)
        raise ValueError( \
                'self and other must have the same sample frequency and sample duration')

    @classmethod
    def concat(cls, arrs, axis=0):
        freqs = set(x.frequency for x in arrs)
        if len(freqs) > 1:
            raise ValueError('all timeseries must have same frequency')

        durations = set(x.duration for x in arrs)
        if len(durations) > 1:
            raise ValueError('all timeseries must have same duration')

        return cls.from_example(np.concatenate(arrs, axis=axis), arrs[0])

    @property
    def samples_per_second(self):
        return int(Picoseconds(int(1e12)) / self.frequency)

    @property
    def duration_in_seconds(self):
        return self.duration / Picoseconds(int(1e12))

    @property
    def samplerate(self):
        return SampleRate(self.frequency, self.duration)

    @property
    def overlap(self):
        return self.samplerate.overlap

    @property
    def span(self):
        overlap = self.duration - self.frequency
        return TimeSlice((len(self) * self.frequency) + overlap)

    @property
    def end(self):
        return self.span.end

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.frequency = getattr(obj, 'frequency', None)
        self.duration = getattr(obj, 'duration', None)

    def __getitem__(self, index):
        if isinstance(index, TimeSlice):
            diff = self.duration - self.frequency
            start_index = \
                max(0, np.floor((index.start - diff) / self.frequency))
            end = self.end if index.duration is None else index.end
            stop_index = np.ceil(end / self.frequency)
            return self[start_index: stop_index]

        return super(ConstantRateTimeSeries, self).__getitem__(index)



from featureflow import Node
import numpy as np
from scipy.fftpack import dct
from scipy.stats.mstats import gmean
from psychacoustics import Chroma as ChromaScale, Bark as BarkScale
from zounds.nputil import safe_log
from zounds.timeseries import ConstantRateTimeSeries, SR44100


class FFT(Node):
    def __init__(self, needs=None, axis=-1):
        super(FFT, self).__init__(needs=needs)
        self._axis = axis

    def _process(self, data):
        transformed = np.fft.fft(data, axis=self._axis)
        sl = [slice(None) for _ in xrange(len(transformed.shape))]
        positive = data.shape[self._axis] // 2
        sl[self._axis] = slice(0, positive, None)
        yield ConstantRateTimeSeries(
                transformed[sl],
                data.frequency,
                data.duration)


class DCT(Node):
    def __init__(self, needs=None, axis=-1):
        super(DCT, self).__init__(needs=needs)
        self._axis = axis

    def _process(self, data):
        yield ConstantRateTimeSeries( \
                dct(data, norm='ortho', axis=self._axis),
                data.frequency,
                data.duration)


# TODO: This constructor should not take a samplerate; that information should
# be encapsulated in the data that's passed in
class Chroma(Node):
    def __init__( \
            self,
            needs=None,
            samplerate=SR44100(),
            nbins=12,
            a440=440.):
        super(Chroma, self).__init__(needs=needs)
        self._nbins = nbins
        self._a440 = a440
        self._samplerate = samplerate
        self._chroma_scale = None

    def _process(self, data):
        data = np.abs(data)
        if self._chroma_scale is None:
            self._chroma_scale = ChromaScale( \
                    self._samplerate.samples_per_second,
                    data.shape[1] * 2,
                    nbands=self._nbins)

        yield ConstantRateTimeSeries( \
                self._chroma_scale.transform(data),
                data.frequency,
                data.duration)


# TODO: This constructor should not take a samplerate; that information should
# be encapsulated in the data that's passed in
class BarkBands(Node):
    def __init__( \
            self,
            needs=None,
            samplerate=SR44100(),
            n_bands=100,
            start_freq_hz=50,
            stop_freq_hz=2e4):
        super(BarkBands, self).__init__(needs=needs)

        self._samplerate = samplerate
        self._n_bands = n_bands
        self._start_freq_hz = start_freq_hz
        self._stop_freq_hz = stop_freq_hz
        self._bark_scale = None

    def _process(self, data):
        data = np.abs(data)
        if self._bark_scale is None:
            self._bark_scale = BarkScale( \
                    self._samplerate.samples_per_second,
                    data.shape[1] * 2,
                    self._n_bands,
                    self._start_freq_hz,
                    self._stop_freq_hz)
        yield ConstantRateTimeSeries( \
                self._bark_scale.transform(data),
                data.frequency,
                data.duration)


class SpectralCentroid(Node):
    """
    Indicates where the "center of mass" of the spectrum is. Perceptually,
    it has a robust connection with the impression of "brightness" of a
    sound.  It is calculated as the weighted mean of the frequencies
    present in the signal, determined using a Fourier transform, with
    their magnitudes as the weights...

    -- http://en.wikipedia.org/wiki/Spectral_centroid
    """
    def __init__(self, needs=None):
        super(SpectralCentroid, self).__init__(needs=needs)

    def _first_chunk(self, data):
        self._bins = np.arange(1, data.shape[-1] + 1)
        self._bins_sum = np.sum(self._bins)
        return data

    def _process(self, data):
        yield ConstantRateTimeSeries(
            (data * self._bins).sum(axis=1) / self._bins_sum,
            data.frequency,
            data.duration)


class SpectralFlatness(Node):
    """
    Spectral flatness or tonality coefficient, also known as Wiener
    entropy, is a measure used in digital signal processing to characterize an
    audio spectrum. Spectral flatness is typically measured in decibels, and
    provides a way to quantify how tone-like a sound is, as opposed to being
    noise-like. The meaning of tonal in this context is in the sense of the
    amount of peaks or resonant structure in a power spectrum, as opposed to
    flat spectrum of a white noise. A high spectral flatness indicates that
    the spectrum has a similar amount of power in all spectral bands - this
    would sound similar to white noise, and the graph of the spectrum would
    appear relatively flat and smooth. A low spectral flatness indicates that
    the spectral power is concentrated in a relatively small number of
    bands - this would typically sound like a mixture of sine waves, and the
    spectrum would appear "spiky"...

    -- http://en.wikipedia.org/wiki/Spectral_flatness
    """
    def __init__(self, needs=None):
        super(SpectralFlatness, self).__init__(needs=needs)

    def _process(self, data):
        m = data.mean(axis=1)
        m[m == 0] = -1e5
        yield ConstantRateTimeSeries(
            gmean(data, axis=1) / m,
            data.frequency,
            data.duration)


class BFCC(Node):
    def __init__(self, needs=None, n_coeffs=13, exclude=1):
        super(BFCC, self).__init__(needs=needs)
        self._n_coeffs = n_coeffs
        self._exclude = exclude

    def _process(self, data):
        data = np.abs(data)
        bfcc = dct(safe_log(data), axis=1) \
            [:, self._exclude: self._exclude + self._n_coeffs]
        yield ConstantRateTimeSeries( \
                bfcc.copy(), data.frequency, data.duration)



import struct

import numpy as np
from featureflow import Node, Feature, Decoder

from zounds.nputil import safe_unit_norm
from zounds.timeseries import ConstantRateTimeSeries, TimeSlice, Picoseconds


class MeasureOfTransience(Node):
    """
    Measure of Transience, as defined in section 5.2.1 of
    http://www.mp3-tech.org/programmer/docs/Masri_thesis.pdf

    Uses the ratio of high-frequency content in the signal to detect onsets.
    Effective for percussive onsets.
    """

    def __init__(self, needs=None):
        super(MeasureOfTransience, self).__init__(needs=needs)

    def _first_chunk(self, data):
        self._bin_numbers = np.arange(1, data.shape[1] + 1)
        padding = np.zeros(data.shape[1])
        padding[:] = 1e-12
        return np.concatenate([padding[None, :], data])

    def _process(self, data):
        magnitude = (data[:, 2:] ** 2)
        energy = magnitude.sum(axis=1)
        hfc = (magnitude * self._bin_numbers[2:]).sum(axis=1)
        energy[energy == 0] = 1e-12
        hfc[hfc == 0] = 1e-12
        mot = (hfc[1:] / hfc[:-1]) * (hfc[1:] / energy[1:])
        yield ConstantRateTimeSeries(mot, data.frequency, data.duration)


class ComplexDomain(Node):
    """
    Complex-domain onset detection as described in
    http://www.eecs.qmul.ac.uk/legacy/dafx03/proceedings/pdfs/dafx81.pdf
    """

    def __init__(self, needs=None):
        super(ComplexDomain, self).__init__(needs=needs)

    def _first_chunk(self, data):
        first = np.zeros((2, 3, data.shape[-1]), dtype=np.complex128)
        first[0, 2:, :] = data[0, :1]
        first[1, 1:, :] = data[0, :2]
        return ConstantRateTimeSeries( \
                np.concatenate([first, data]),
                data.frequency,
                data.duration)

    def _process(self, data):
        # delta between expected and actual phase
        # TODO: unwrap phases before computing deltas, to avoid artifacts
        # or discontinuties from phase boundary wrapping
        angle = np.angle(data)
        angle = np.unwrap(angle, axis=1)
        angle = np.angle(angle[:, 2] - (2 * angle[:, 1]) + angle[:, 0])

        # expected magnitude
        expected = np.abs(data[:, 1, :])
        # actual magnitude
        actual = np.abs(data[:, 2, :])
        # detection function array
        detect = np.zeros(angle.shape)

        # where phase delta is zero, detection function is the difference 
        # between expected and actual magnitude
        zero_phase_delta_indices = np.where(angle == 0)
        detect[zero_phase_delta_indices] = \
            (expected - actual)[zero_phase_delta_indices]

        # where phase delta is non-zero, detection function combines magnitude
        # and phase deltas
        nonzero_phase_delta_indices = np.where(angle != 0)
        detect[nonzero_phase_delta_indices] = (
            ((expected ** 2) + (actual ** 2) -
             (2 * expected * actual * np.cos(angle))) ** 0.5)[
            nonzero_phase_delta_indices]

        # TODO: This duration isn't right.  It should probably be 
        # data.duration // 3
        yield ConstantRateTimeSeries( \
                detect.sum(axis=1),
                data.frequency,
                data.duration)


class Flux(Node):
    def __init__(self, unit_norm=False, needs=None):
        super(Flux, self).__init__(needs=needs)
        self._memory = None
        self._unit_norm = unit_norm

    def _enqueue(self, data, pusher):
        if self._memory is None:
            self._cache = np.vstack((data[0], data))
        else:
            self._cache = np.vstack((self._memory, data))
        self._cache = ConstantRateTimeSeries( \
                self._cache,
                data.frequency,
                data.duration)
        self._memory = data[-1]

    def _process(self, data):
        if self._unit_norm:
            data = safe_unit_norm(data)
        diff = np.diff(data, axis=0)
        yield ConstantRateTimeSeries( \
                np.linalg.norm(diff, axis=-1),
                data.frequency,
                data.duration)


class BasePeakPicker(Node):
    def __init__(self, needs=None):
        super(BasePeakPicker, self).__init__(needs=needs)
        self._pos = Picoseconds(0)

    def _onset_indices(self, data):
        raise NotImplementedError()

    def _last_chunk(self):
        yield self._pos

    def _process(self, data):
        if self._pos == Picoseconds(0):
            yield self._pos

        indices = self._onset_indices(data)
        timestamps = self._pos + (indices * data.frequency)
        self._pos += len(data) * data.frequency
        yield timestamps


class MovingAveragePeakPicker(BasePeakPicker):
    def __init__(self, aggregate=np.mean, needs=None):
        super(MovingAveragePeakPicker, self).__init__(needs=needs)
        self._aggregate = aggregate

    def _first_chunk(self, data):
        self._center = data.shape[1] // 2
        return data

    def _onset_indices(self, data):
        agg = self._aggregate(data, axis=1) * 1.25
        # find indices that are peaks
        diff = np.diff(data[:, self._center - 2: self._center + 1])
        peaks = (diff[:, 0] > 0) & (diff[:, 1] < 0)
        # find indices that are above the local average or median
        over_thresh = data[:, self._center] > agg
        # return the intersection of the two
        return np.where(peaks & over_thresh)[0]


class SparseTimestampEncoder(Node):
    content_type = 'application/octet-stream'

    def __init__(self, needs=None):
        super(SparseTimestampEncoder, self).__init__(needs=needs)
        # TODO: Add a class (mixin) in the flow library for this pattern where
        # the _process implementarion changes depending on whether it's the first
        # call or a subsequent one
        self._initialized = False

    def _process(self, data):
        if not self._initialized:
            sd = str(data.dtype)
            yield struct.pack('B', len(sd))
            yield sd
            self._initialized = True

        yield data.astype(np.uint64).tostring()


# TODO: Encode/decode tests
# TODO: A subclass of this that turns each pair into a timeslice
# TODO: Should PeakPicker always emit the *end* of the timeseries, so that the
# final timeslice can be produced correctly?
class SparseTimestampDecoder(Decoder):
    def __init__(self):
        super(SparseTimestampDecoder, self).__init__()

    def __call__(self, flo):
        dtype_len = struct.unpack('B', flo.read(1))[0]
        dtype = np.dtype(flo.read(dtype_len))
        data = np.fromstring(flo.read(), dtype=np.uint64)
        return np.array(data, dtype=dtype)

    def __iter__(self, flo):
        yield self(flo)


class TimeSliceDecoder(SparseTimestampDecoder):
    def __init__(self):
        super(TimeSliceDecoder, self).__init__()

    def __call__(self, flo):
        timestamps = super(TimeSliceDecoder, self).__call__(flo)
        durations = np.diff(timestamps)
        return [TimeSlice(d, s) for s, d in zip(timestamps, durations)]

    def __iter__(self, flo):
        yield self(flo)


class SparseTimestampFeature(Feature):
    def __init__( \
            self,
            extractor,
            needs=None,
            store=False,
            key=None,
            encoder=SparseTimestampEncoder,
            decoder=SparseTimestampDecoder(),
            **extractor_args):
        super(SparseTimestampFeature, self).__init__( \
                extractor,
                needs=needs,
                store=store,
                encoder=encoder,
                decoder=decoder,
                key=key,
                **extractor_args)


class TimeSliceFeature(Feature):
    def __init__( \
            self,
            extractor,
            needs=None,
            store=False,
            key=None,
            encoder=SparseTimestampEncoder,
            decoder=TimeSliceDecoder(),
            **extractor_args):
        super(TimeSliceFeature, self).__init__( \
                extractor,
                needs=needs,
                store=store,
                encoder=encoder,
                decoder=decoder,
                key=key,
                **extractor_args)
